import sys
import unittest
from unittest.mock import Mock

from AnyQt.QtTest import QSignalSpy

from orangecanvas.scheme import Scheme, SchemeNode, SchemeLink
from orangecanvas.scheme.signalmanager import (
    SignalManager, Signal, compress_signals, compress_single, LazyValue
)
from orangecanvas.registry import tests as registry_tests
from orangecanvas.gui.test import QCoreAppTestCase


class TestLazyValue(unittest.TestCase):
    def test_singletonnes(self):
        i1 = LazyValue[int]
        i2 = LazyValue[int]
        f1 = LazyValue[float]
        self.assertIs(i1, i2)
        self.assertIsNot(i1, f1)
        self.assertIsNot(i2, f1)

    def test_repr(self):
        self.assertEqual(repr(LazyValue[int]), "LazyValue[int]")

    def test_get_value_and_cached(self):
        f = Mock(return_value=42)

        lazy = LazyValue[int](f)
        f.assert_not_called()
        self.assertFalse(lazy.is_cached)

        self.assertEqual(lazy.get_value(), 42)
        f.assert_called_once()
        self.assertTrue(lazy.is_cached)

        self.assertEqual(lazy.get_value(), 42)
        f.assert_called_once()
        self.assertTrue(lazy.is_cached)

    def test_type(self):
        self.assertIs(LazyValue[int].type(), int)
        self.assertIs(LazyValue[float].type(), float)

    def test_release_closure(self):
        deleted = Mock()

        def commit():
            class S:
                def __del__(self):
                    deleted()

            s = S()

            def f():
                return 42 + bool(s)
            return LazyValue[int](f)

        lazy = commit()
        deleted.assert_not_called()
        lazy.get_value()
        deleted.assert_called_once()

    def test_interrupt(self):
        interrupt = Mock()
        lazy = LazyValue[int](Mock(), interrupt)
        del lazy
        interrupt.assert_called_once()

    def test_extra_args(self):
        lazy = LazyValue[int](Mock(), a=1, b=2)
        self.assertEqual(lazy.a, 1)
        self.assertEqual(lazy.b, 2)


class TestingSignalManager(SignalManager):
    def is_blocking(self, node):
        return bool(node.property("-blocking"))

    def send_to_node(self, node, signals):
        for sig in signals:
            name = sig.link.sink_channel.name
            node.setProperty("-input-" + name, sig.value)

        for out in node.description.outputs:
            self.send(node, out, "hello")


class TestSignalManager(QCoreAppTestCase):
    def setUp(self):
        super().setUp()
        reg = registry_tests.small_testing_registry()
        scheme = Scheme()
        zero = scheme.new_node(reg.widget("zero"))
        one = scheme.new_node(reg.widget("one"))
        add = scheme.new_node(reg.widget("add"))
        scheme.new_link(zero, "value", add, "left")
        scheme.new_link(one, "value", add, "right")
        sm = TestingSignalManager()
        sm.set_workflow(scheme)
        sm.start()
        self.reg = reg
        self.scheme = scheme
        self.signal_manager = sm

    def test(self):
        workflow = self.scheme
        sm = TestingSignalManager()
        sm.set_workflow(workflow)
        sm.set_workflow(workflow)
        sm.set_workflow(None)
        sm.set_workflow(workflow)
        sm.start()

        self.assertFalse(sm.has_pending())

        sm.stop()
        sm.pause()
        sm.resume()
        n0, n1, n3 = workflow.nodes

        sm.send(n0, n0.description.outputs[0], 'hello')
        sm.send(n1, n1.description.outputs[0], 'hello')
        spy = QSignalSpy(sm.processingFinished[SchemeNode])
        self.assertTrue(spy.wait())
        self.assertSequenceEqual(list(spy), [[n3]])
        self.assertEqual(n3.property("-input-left"), 'hello')
        self.assertEqual(n3.property("-input-right"), 'hello')

        self.assertFalse(sm.has_pending())
        workflow.remove_link(workflow.links[0])
        self.assertTrue(sm.has_pending())

        spy = QSignalSpy(sm.processingFinished[SchemeNode])
        self.assertTrue(spy.wait())

        self.assertEqual(n3.property("-input-left"), None)
        self.assertEqual(n3.property("-input-right"), 'hello')

    def test_add_link_disabled(self):
        workflow = self.scheme
        sm = self.signal_manager
        n0, n1, n2 = workflow.nodes
        l0, l1 = workflow.links
        workflow.remove_link(l0)
        sm.send(n0, n0.description.outputs[0], 1)
        sm.send(n1, n1.description.outputs[0], 2)
        sm.process_queued()

        self.assertFalse(sm.has_pending())
        l0.set_enabled(False)
        workflow.insert_link(0, l0)
        self.assertSequenceEqual(
            sm.pending_input_signals(n2), [Signal.New(l0, None, None, 0)]
        )
        l0.set_enabled(True)
        self.assertSequenceEqual(
            sm.pending_input_signals(n2),
            [
                Signal.New(l0, None, None, 0),
                Signal.Update(l0, 1, None, 0),
            ]
        )

    def test_link_new_dispatch_after_enable(self):
        workflow = self.scheme
        sm = self.signal_manager
        n0, n1, n2 = workflow.nodes
        l0, l1 = workflow.links
        l0.set_enabled(False)
        sm.send(n0, n0.description.outputs[0], -42)
        sm.send(n0, n0.description.outputs[0], 42)
        self.assertSequenceEqual(
            sm.pending_input_signals(n2), [Signal.New(l0, None, None, 0)]
        )
        sm.process_queued()
        l0.set_enabled(True)
        self.assertSequenceEqual(
            sm.pending_input_signals(n2), [Signal.Update(l0, 42, None, 0)]
        )
        sm.process_queued()

    def test_invalidated_flags(self):
        workflow = self.scheme
        sm = self.signal_manager
        n0, n1, n2 = workflow.nodes[:3]
        l0, l1 = workflow.links[:2]

        self.assertFalse(l0.runtime_state() & SchemeLink.Invalidated)
        sm.send(n0, n0.description.outputs[0], 'hello')
        self.assertFalse(l0.runtime_state() & SchemeLink.Invalidated)
        self.assertIn(n2, sm.node_update_front())

        sm.invalidate(n0, n0.description.outputs[0])
        self.assertTrue(l0.runtime_state() & SchemeLink.Invalidated)
        self.assertTrue(sm.has_invalidated_outputs(n0))
        self.assertTrue(sm.has_invalidated_inputs(n2))
        self.assertNotIn(n2, sm.node_update_front())

        sm.send(n0, n0.description.outputs[0], 'hello')
        self.assertFalse(l0.runtime_state() & SchemeLink.Invalidated)
        self.assertFalse(sm.has_invalidated_outputs(n0))
        self.assertFalse(sm.has_invalidated_inputs(n2))
        self.assertIn(n2, sm.node_update_front())

        sm.invalidate(n1, n1.description.outputs[0])
        n3 = workflow.new_node(self.reg.widget('add'))
        l2 = workflow.new_link(
            n1, n1.output_channel('value'), n3, n3.input_channel('left')
        )

        self.assertTrue(l2.test_runtime_state(SchemeLink.Invalidated))
        self.assertTrue(sm.has_invalidated_inputs(n3))
        self.assertNotIn(n3, sm.node_update_front())

        workflow.remove_link(l2)
        self.assertFalse(sm.has_invalidated_inputs(n3))
        self.assertNotIn(n3, sm.node_update_front())

        # invalidated must not propagate via disabled links
        self.assertNotIn(n2, sm.node_update_front())
        l1.set_enabled(False)
        self.assertIn(n2, sm.node_update_front())
        self.assertFalse(sm.has_invalidated_inputs(n2))
        l1.set_enabled(True)
        self.assertNotIn(n2, sm.node_update_front())
        self.assertTrue(sm.has_invalidated_inputs(n2))

    def test_pending_flags(self):
        workflow = self.scheme
        sm = self.signal_manager
        n0, n1, n3 = workflow.nodes[:3]
        l0, l1 = workflow.links[:2]

        self.assertFalse(n3.test_state_flags(SchemeNode.Pending))
        self.assertFalse(l0.runtime_state() & SchemeLink.Pending)
        sm.send(n0, n0.description.outputs[0], 'hello')
        self.assertTrue(n3.test_state_flags(SchemeNode.Pending))
        self.assertTrue(l0.runtime_state() & SchemeLink.Pending)

        spy = QSignalSpy(sm.processingFinished)
        assert spy.wait()

        self.assertFalse(n3.test_state_flags(SchemeNode.Pending))
        self.assertFalse(l0.runtime_state() & SchemeLink.Pending)

    def test_ready_flags(self):
        workflow = self.scheme
        sm = self.signal_manager
        n0, n1, n3 = workflow.nodes[:3]
        sm.send(n0, n0.output_channel("value"), 'hello')
        sm.send(n1, n1.output_channel("value"), 'hello')
        self.assertIn(n3, sm.node_update_front())
        n3.set_state_flags(SchemeNode.NotReady, True)
        spy = QSignalSpy(sm.processingStarted[SchemeNode])
        sm.process_next()
        self.assertNotIn([n3], list(spy))
        n3.set_state_flags(SchemeNode.NotReady, False)
        assert spy.wait()
        self.assertIn([n3], list(spy))

    def test_start_finished(self):
        workflow = self.scheme
        sm = self.signal_manager
        n0, n1, n3 = workflow.nodes[:3]
        start_spy = QSignalSpy(sm.started)
        fin_spy = QSignalSpy(sm.finished)
        sm.send(n0, n0.output_channel("value"), 'hello')
        sm.send(n1, n1.output_channel("value"), 'hello')
        assert fin_spy.wait()
        self.assertEqual(len(start_spy), 1)
        self.assertEqual(len(fin_spy), 1)
        sm.send(n1, n1.output_channel("value"), 'hello')
        assert fin_spy.wait()
        self.assertEqual(len(start_spy), 2)
        self.assertEqual(len(fin_spy), 2)

    def test_compress_signals(self):
        workflow = self.scheme
        link = workflow.links[0]
        self.assertSequenceEqual(compress_signals([]), [])
        signals_in = [
            Signal(link, 1, None),
            Signal(link, 3, None),
            Signal(link, 2, None),
        ]
        self.assertSequenceEqual(
            compress_signals(signals_in),
            signals_in[-1:]
        )
        signals_in = [
            Signal(link, None, None),
            Signal(link, 3, None),
            Signal(link, 2, None),
        ]
        self.assertSequenceEqual(
            compress_signals(signals_in),
            [signals_in[0], signals_in[-1]]
        )
        signals_in = [
            Signal(link, None, 1),
            Signal(link, 3, 1),
            Signal(link, 2, 2),
        ]
        self.assertSequenceEqual(
            compress_signals(signals_in),
            signals_in,
        )
        signals_in = [
            Signal(link, 1, 1),
            Signal(link, None, 1),
            Signal(link, 2, 2),
        ]
        self.assertSequenceEqual(
            compress_signals(signals_in),
            signals_in[1:],
        )
        signals_in = [
            Signal(link, None, 1),
            Signal(link, None, 1),
        ]
        self.assertSequenceEqual(
            compress_signals(signals_in),
            signals_in[1:],
        )

    def test_compress_signals_single(self):
        New, Update, Close = Signal.New, Signal.Update, Signal.Close
        workflow = self.scheme
        link = workflow.links[0]
        self.assertSequenceEqual(
            compress_single([]), []
        )
        signals = [Update(link, None, 1)]
        self.assertSequenceEqual(
            compress_single(signals), signals
        )
        signals = [Update(link, None, 1), Update(link, 1, 1)]
        self.assertSequenceEqual(
            compress_single(signals), signals
        )
        signals = [Update(link, 1, 1), Update(link, None, 1)]
        self.assertSequenceEqual(
            compress_single(signals), [signals[-1]]
        )
        signals = [
            Update(link, None, 1),
            Update(link, 1, 1),
            Update(link, None, 1),
        ]
        self.assertSequenceEqual(
            compress_single(signals),
            [signals[-1]]
        )
        signals = [
            Update(link, None, 1),
            Update(link, 1, 1),
            Update(link, 2, 1),
        ]
        self.assertSequenceEqual(
            compress_single(signals),
            [signals[0], signals[-1]]
        )
        signals = [New(link, None, 1), Close(link, None, 1)]
        self.assertSequenceEqual(
            compress_single(signals), signals,
        )
        signals = [
            New(link, 1, 1),
            Update(link, 2, 1),
            Close(link, None, 1)
        ]
        self.assertSequenceEqual(
            compress_single(signals), [signals[0], signals[-1]]
        )
        signals = [
            New(link, 1, 1),
            Update(link, 1, 1),
            Close(link, None, 1),
            New(link, 1, 1)
        ]
        self.assertSequenceEqual(
            compress_single(signals),
            [signals[0], *signals[2:]]
        )
        signals = [
            Update(link, 1, 1),
            Update(link, 2, 1),
            Close(link, None, 1)
        ]
        self.assertSequenceEqual(
            compress_single(signals), [signals[-1]]
        )
        signals = [
            Update(link, 1, 1),
            Update(link, None, 1),
            Update(link, 2, 1),
            Close(link, None, 1)
        ]
        self.assertSequenceEqual(
            compress_single(signals), [signals[-1]],
        )
        signals = [
            Update(link, 1, 1),
            Update(link, 2, 1),
            Close(link, None, 1),
        ]
        self.assertSequenceEqual(
            compress_single(signals), [signals[-1]]
        )
        signals = [
            Update(link, 1, 1),
            Update(link, 2, 1),
            Close(link, None, 1),
            New(link, None, 1),
        ]
        self.assertSequenceEqual(
            compress_single(signals), signals[-2:]
        )

    def test_compress_signals_typed(self):
        l1, l2 = self.scheme.links[0], self.scheme.links[1]
        New, Update, Close = Signal.New, Signal.Update, Signal.Close
        signals = [
            New(l1, 1, index=0),
            Update(l1, 2, index=0),
            New(l2, "a", index=0),
            Update(l2, 2, index=0),
            Close(l1, None, index=1),
            New(l1, None, index=1),
            Update(l2, "b", index=0)
        ]
        # must preserve relative order of New/Close
        self.assertSequenceEqual(
            compress_signals(signals),
            [
                New(l1, 1, index=0),
                New(l2, "a", index=0),
                Close(l1, None, index=1),
                New(l1, None, index=1),
                Update(l2, "b", index=0)
            ],
        )
        signals = [
            Update(l1, 2, index=0),
            New(l2, "a", index=0),
            Update(l2, 2, index=0),
            Close(l1, None, index=1),
        ]
        self.assertSequenceEqual(
            compress_signals(signals),
            [
                New(l2, "a", index=0),
                Update(l2, 2, index=0),
                Close(l1, None, index=1),
            ],
        )



"""
Parsers for intersphinx inventory files

Taken from `sphinx.ext.intersphinx`

"""

import re
import codecs
import zlib

b = str
UTF8StreamReader = codecs.lookup('utf-8')[2]


def read_inventory_v1(f, uri, join):
    f = UTF8StreamReader(f)
    invdata = {}
    line = f.next()
    projname = line.rstrip()[11:]
    line = f.next()
    version = line.rstrip()[11:]
    for line in f:
        name, type, location = line.rstrip().split(None, 2)
        location = join(uri, location)
        # version 1 did not add anchors to the location
        if type == 'mod':
            type = 'py:module'
            location += '#module-' + name
        else:
            type = 'py:' + type
            location += '#' + name
        invdata.setdefault(type, {})[name] = (projname, version, location, '-')
    return invdata


def read_inventory_v2(f, uri, join, bufsize=16*1024):
    invdata = {}
    line = f.readline()
    projname = line.rstrip()[11:].decode('utf-8')
    line = f.readline()
    version = line.rstrip()[11:].decode('utf-8')
    line = f.readline().decode('utf-8')
    if 'zlib' not in line:
        raise ValueError

    def read_chunks():
        decompressor = zlib.decompressobj()
        for chunk in iter(lambda: f.read(bufsize), b''):
            yield decompressor.decompress(chunk)
        yield decompressor.flush()

    def split_lines(chunkiter):
        buf = b''
        for chunk in chunkiter:
            buf += chunk
            lineend = buf.find(b'\n')
            while lineend != -1:
                yield buf[:lineend].decode('utf-8')
                buf = buf[lineend + 1:]
                lineend = buf.find(b'\n')
        assert not buf

    for line in split_lines(read_chunks()):
        # be careful to handle names with embedded spaces correctly
        m = re.match(r'(?x)(.+?)\s+(\S*:\S*)\s+(\S+)\s+(\S+)\s+(.*)',
                     line.rstrip())
        if not m:
            continue
        name, type, prio, location, dispname = m.groups()
        if location.endswith('$'):
            location = location[:-1] + name
        location = join(uri, location)
        invdata.setdefault(type, {})[name] = (projname, version,
                                              location, dispname)
    return invdata



from typing import Sequence

import numpy as np

from AnyQt.QtGui import QImage, QColor


def qrgb(
        r: Sequence[int], g: Sequence[int], b: Sequence[int]
) -> Sequence[int]:
    """A vectorized `qRgb`."""
    r, g, b = map(lambda a: np.asarray(a, dtype=np.uint32), (r, g, b))
    return (0xff << 24) | ((r & 0xff) << 16) | ((g & 0xff) << 8) | (b & 0xff)


def qrgba(
        r: Sequence[int], g: Sequence[int], b: Sequence[int], a: Sequence[int]
) -> Sequence[int]:
    """A vectorized `qRgba`."""
    r, g, b, a = map(lambda a: np.asarray(a, dtype=np.uint32), (r, g, b, a))
    return ((a & 0xff) << 24) | ((r & 0xff) << 16) | ((g & 0xff) << 8) | (b & 0xff)


def qgray(
        r: Sequence[int], g: Sequence[int], b: Sequence[int]
) -> Sequence[int]:
    """A vectorized `qGray`."""
    r, g, b = map(lambda a: np.asarray(a, dtype=np.uint16), (r, g, b))
    return (r * 11 + g * 16 + b * 5) // 32


def qred(rgb: Sequence[int]) -> Sequence[int]:
    """A vectorized `qRed`."""
    rgb = np.asarray(rgb, np.uint32)
    return (rgb >> 16) & 0xff


def qgreen(rgb: Sequence[int]) -> Sequence[int]:
    """A vectorized `qGreen`."""
    rgb = np.asarray(rgb, np.uint32)
    return (rgb >> 8) & 0xff


def qblue(rgb: Sequence[int]) -> Sequence[int]:
    """A vectorized `qBlue`."""
    rgb = np.asarray(rgb, np.uint32)
    return rgb & 0xff


def qalpha(rgb: Sequence[int]) -> Sequence[int]:
    """A vectorized `qAlpha`."""
    rgb = np.asarray(rgb, np.uint32)
    return (rgb >> 24) & 0xff


def grayscale_invert(
        src: QImage, foreground: QColor, background: QColor
) -> QImage:
    """
    Convert the `src` image to grayscale and invert it into background
    to foreground (gray) range.

    Parameters
    ----------
    src: QImage
    foreground: QColor
    background: QColor

    Returns
    -------
    image: QImage
    """
    image = src.convertToFormat(QImage.Format_ARGB32)
    size = image.size()
    w, h = shape = size.width(), size.height()
    buffer = image.bits().asarray(w * h * 4)
    view = np.frombuffer(buffer, np.uint32).reshape(shape)
    r, g, b, a = qred(view), qgreen(view), qblue(view), qalpha(view)
    gray = qgray(r, g, b)
    factor = gray / 255
    foreground = qgray(foreground.red(), foreground.blue(), foreground.green())
    background = qgray(background.red(), background.blue(), background.green())
    if foreground > background:
        minv_, maxv_ = background, foreground
    else:
        minv_, maxv_ = foreground, background
    inv = (1 - factor) * (maxv_ - minv_) + minv_
    inv = np.asarray(inv, np.uint8)
    rgba = qrgba(inv, inv, inv, a)
    res = QImage(w, h, QImage.Format_ARGB32)
    return qimage_copy_from_buffer(res, rgba)


def qimage_copy_from_buffer(image: QImage, data: np.ndarray) -> QImage:
    """
    Copy the `data` to `image`.

    Parameters
    ----------
    image: QImage
        The destination image.
    data: np.ndarray
        The raw source data in the same format as the `image`.
    """
    w, h = image.width(), image.height()
    if data.shape != (w, h):
        raise ValueError(
            f"Wrong data.shape (expected ({w}, {h}) got {data.shape})"
        )
    d = image.depth() // 8
    dtype = {
        1: np.uint8, 2: np.uint16, 4: np.uint32, 8: np.uint64
    }[d]
    dest = image.bits().asarray(w * h * d)
    dest = np.frombuffer(dest, dtype).reshape((w, h))
    dest[:] = data
    return image



"""
"""
import io
import sys
import warnings
import traceback
from types import TracebackType
from typing import Any, Optional, List, Type, Iterable, Tuple, Union, Mapping

from AnyQt.QtWidgets import (
    QWidget, QPlainTextEdit, QVBoxLayout, QSizePolicy, QPlainTextDocumentLayout
)
from AnyQt.QtGui import (
    QTextCursor, QTextCharFormat, QTextOption, QFontDatabase, QTextDocument,
    QTextDocumentFragment
)
from AnyQt.QtCore import Qt, QObject, QCoreApplication, QThread, QSize
from AnyQt.QtCore import pyqtSignal as Signal, pyqtSlot as Slot

from orangecanvas.gui.utils import update_char_format
from orangecanvas.utils import findf


class TerminalView(QPlainTextEdit):
    def __init__(self, *args, **kwargs):
        # type: (Any, Any) -> None
        super().__init__(*args, **kwargs)
        self.setFrameStyle(QPlainTextEdit.NoFrame)
        self.setTextInteractionFlags(Qt.TextBrowserInteraction)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        self.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        font = QFontDatabase.systemFont(QFontDatabase.FixedFont)
        self.setFont(font)
        self.setAttribute(Qt.WA_SetFont, False)

    def sizeHint(self):
        # type: () -> QSize
        metrics = self.fontMetrics()
        width = metrics.boundingRect("X" * 81).width()
        height = metrics.lineSpacing()
        scroll_width = self.verticalScrollBar().width()
        size = QSize(width + scroll_width, height * 25)
        return size


class TerminalTextDocument(QTextDocument):
    def __init__(self, parent=None, **kwargs):
        super().__init__(parent, **kwargs)
        self.setDocumentLayout(QPlainTextDocumentLayout(self))
        self.__currentCharFormat = QTextCharFormat()
        if 'defaultFont' not in kwargs:
            defaultFont = QFontDatabase.systemFont(QFontDatabase.FixedFont)
            self.setDefaultFont(defaultFont)
        self.__streams = []

    def setCurrentCharFormat(self, charformat: QTextCharFormat) -> None:
        """Set the QTextCharFormat to be used when writing."""
        assert QThread.currentThread() is self.thread()
        if self.__currentCharFormat != charformat:
            self.__currentCharFormat = QTextCharFormat(charformat)

    def currentCharFormat(self) -> QTextCharFormat:
        """Return the current char format."""
        return QTextCharFormat(self.__currentCharFormat)

    def textCursor(self) -> QTextCursor:
        """Return a text cursor positioned at the end of the document."""
        cursor = QTextCursor(self)
        cursor.movePosition(QTextCursor.End, QTextCursor.MoveAnchor)
        cursor.setCharFormat(self.__currentCharFormat)
        return cursor

    # ----------------------
    # A file like interface.
    # ----------------------

    @Slot(str)
    def write(self, string: str) -> None:
        assert QThread.currentThread() is self.thread()
        cursor = self.textCursor()
        cursor.insertText(string)

    @Slot(object)
    def writelines(self, lines: Iterable[str]) -> None:
        assert QThread.currentThread() is self.thread()
        self.write("".join(lines))

    @Slot()
    def flush(self) -> None:
        assert QThread.currentThread() is self.thread()

    def writeWithFormat(self, string: str, charformat: QTextCharFormat) -> None:
        assert QThread.currentThread() is self.thread()
        cursor = self.textCursor()
        cursor.setCharFormat(charformat)
        cursor.insertText(string)

    def writelinesWithFormat(self, lines, charformat):
        # type: (List[str], QTextCharFormat) -> None
        self.writeWithFormat("".join(lines), charformat)

    def formatted(self, color=None, background=None, weight=None,
                  italic=None, underline=None, font=None):
        # type: (...) -> Formatter
        """
        Return a formatted file like object proxy.
        """
        charformat = update_char_format(
            self.currentCharFormat(), color, background, weight,
            italic, underline, font
        )
        return Formatter(self, charformat)

    __streams: List[Tuple['TextStream', Optional['Formatter']]]

    def connectedStreams(self) -> List['TextStream']:
        """Return all streams connected using `connectStream`."""
        return [s for s, _ in self.__streams]

    def connectStream(
            self, stream: 'TextStream',
            charformat: Optional[QTextCharFormat] = None,
            **kwargs
    ) -> None:
        """
        Connect a :class:`TextStream` instance to this document.

        The `stream` connection will be 'inherited' by `clone()`
        """
        if kwargs and charformat is not None:
            raise TypeError("'charformat' and kwargs cannot be used together")
        if kwargs:
            charformat = update_char_format(QTextCharFormat(), **kwargs)
        writer: Optional[Formatter] = None
        if charformat is not None:
            writer = Formatter(self, charformat)
        self.__streams.append((stream, writer))
        if writer is not None:
            stream.stream.connect(writer.write)
        else:
            stream.stream.connect(self.write)

    def disconnectStream(self, stream: 'TextStream'):
        """
        Disconnect a :class:`TextStream` instance from this document.
        """
        item = findf(self.__streams, lambda t: t[0] is stream)
        if item is not None:
            self.__streams.remove(item)
            _, writer = item
            if writer is not None:
                stream.stream.disconnect(writer.write)
            else:
                stream.stream.disconnect(self.write)

    def clone(self, parent=None) -> 'TerminalTextDocument':
        """Create a new TerminalTextDocument that is a copy of this document."""
        clone = type(self)()
        clone.setParent(parent)
        clone.setDocumentLayout(QPlainTextDocumentLayout(clone))
        cursor = QTextCursor(clone)
        cursor.insertFragment(QTextDocumentFragment(self))
        clone.rootFrame().setFrameFormat(self.rootFrame().frameFormat())
        clone.setDefaultStyleSheet(self.defaultStyleSheet())
        clone.setDefaultFont(self.defaultFont())
        clone.setDefaultTextOption(self.defaultTextOption())
        clone.setCurrentCharFormat(self.currentCharFormat())
        for s, w in self.__streams:
            clone.connectStream(s, w.charformat if w is not None else None)
        return clone


class OutputView(QWidget):
    def __init__(self, parent=None, **kwargs):
        # type: (Optional[QWidget], Any) -> None
        super().__init__(parent, **kwargs)

        self.__lines = 5000

        self.setLayout(QVBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)

        self.__text = TerminalView()
        self.__text.setDocument(TerminalTextDocument(self.__text))
        self.__text.setWordWrapMode(QTextOption.NoWrap)
        self.__text.setMaximumBlockCount(self.__lines)

        self.layout().addWidget(self.__text)

    def setMaximumLines(self, lines):
        # type: (int) -> None
        """
        Set the maximum number of lines to keep displayed.
        """
        if self.__lines != lines:
            self.__lines = lines
            self.__text.setMaximumBlockCount(lines)

    def maximumLines(self):
        # type: () -> int
        """
        Return the maximum number of lines in the display.
        """
        return self.__lines

    def clear(self):
        # type: () -> None
        """
        Clear the displayed text.
        """
        assert QThread.currentThread() is self.thread()
        self.__text.clear()

    def setCurrentCharFormat(self, charformat):
        # type: (QTextCharFormat) -> None
        """Set the QTextCharFormat to be used when writing.
        """
        assert QThread.currentThread() is self.thread()
        self.document().setCurrentCharFormat(charformat)

    def currentCharFormat(self):
        # type: () -> QTextCharFormat
        return QTextCharFormat(self.document().currentCharFormat())

    def toPlainText(self):
        # type: () -> str
        """
        Return the full contents of the output view.
        """
        return self.__text.toPlainText()

    # A file like interface.
    @Slot(str)
    def write(self, string):
        # type: (str) -> None
        assert QThread.currentThread() is self.thread()
        doc = self.document()
        doc.write(string)

    @Slot(object)
    def writelines(self, lines):
        # type: (List[str]) -> None
        assert QThread.currentThread() is self.thread()
        self.write("".join(lines))

    @Slot()
    def flush(self):
        # type: () -> None
        assert QThread.currentThread() is self.thread()

    def writeWithFormat(self, string, charformat):
        # type: (str, QTextCharFormat) -> None
        assert QThread.currentThread() is self.thread()
        doc = self.document()
        doc.writeWithFormat(string, charformat)

    def writelinesWithFormat(self, lines, charformat):
        # type: (List[str], QTextCharFormat) -> None
        assert QThread.currentThread() is self.thread()
        self.writeWithFormat("".join(lines), charformat)

    def formatted(self, color=None, background=None, weight=None,
                  italic=None, underline=None, font=None):
        # type: (...) -> Formatter
        """
        Return a formatted file like object proxy.
        """
        charformat = update_char_format(
            self.currentCharFormat(), color, background, weight,
            italic, underline, font
        )
        return Formatter(self, charformat)

    def document(self) -> TerminalTextDocument:
        return self.__text.document()

    def setDocument(self, document: TerminalTextDocument) -> None:
        document.setMaximumBlockCount(self.__lines)
        document.setDefaultFont(self.__text.font())
        self.__text.setDocument(document)

    def formated(self, *args, **kwargs):
        warnings.warn(
            "'Use 'formatted'", DeprecationWarning, stacklevel=2
        )
        return self.formatted(*args, **kwargs)


class Formatter(QObject):
    def __init__(self, outputview, charformat):
        # type: (Union[TerminalTextDocument, OutputView], QTextCharFormat) -> None
        # Parent to the output view. Ensure the formatter does not outlive it.
        super().__init__(outputview)
        self.outputview = outputview
        self.charformat = charformat

    @Slot(str)
    def write(self, string):
        # type: (str) -> None
        self.outputview.writeWithFormat(string, self.charformat)

    @Slot(object)
    def writelines(self, lines):
        # type: (List[str]) -> None
        self.outputview.writelinesWithFormat(lines, self.charformat)

    @Slot()
    def flush(self):
        # type: () -> None
        self.outputview.flush()

    def formatted(self, color=None, background=None, weight=None,
                  italic=None, underline=None, font=None):
        # type: (...) -> Formatter
        charformat = update_char_format(self.charformat, color, background,
                                        weight, italic, underline, font)
        return Formatter(self.outputview, charformat)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.outputview = None
        self.charformat = None
        self.setParent(None)

    def formated(self, *args, **kwargs):
        warnings.warn(
            "Use 'formatted'", DeprecationWarning, stacklevel=2
        )
        return self.formatted(*args, **kwargs)


class formater(Formatter):
    def __init__(self, *args, **kwargs):
        warnings.warn(
            "Deprecated: Renamed to Formatter.",
            DeprecationWarning, stacklevel=2
        )
        super().__init__(*args, **kwargs)


class TextStream(QObject):
    stream = Signal(str)
    flushed = Signal()
    __closed = False

    def close(self):
        # type: () -> None
        self.__closed = True

    def closed(self):
        # type: () -> bool
        return self.__closed

    def isatty(self):
        # type: () -> bool
        return False

    def write(self, string):
        # type: (str) -> None
        if self.__closed:
            raise ValueError("write operation on a closed stream.")
        self.stream.emit(string)

    def writelines(self, lines):
        # type: (List[str]) -> None
        if self.__closed:
            raise ValueError("write operation on a closed stream.")
        self.stream.emit("".join(lines))

    def flush(self):
        # type: () -> None
        if self.__closed:
            raise ValueError("write operation on a closed stream.")
        self.flushed.emit()

    def writeable(self):
        # type: () -> bool
        return True

    def readable(self):
        # type: () -> bool
        return False

    def seekable(self):
        # type: () -> bool
        return False

    encoding = None
    errors = None
    newlines = None
    buffer = None

    def detach(self):
        raise io.UnsupportedOperation("detach")

    def read(self, size=-1):
        raise io.UnsupportedOperation("read")

    def readline(self, size=-1):
        raise io.UnsupportedOperation("readline")

    def readlines(self):
        raise io.UnsupportedOperation("readlines")

    def fileno(self):
        raise io.UnsupportedOperation("fileno")

    def seek(self, offset, whence=io.SEEK_SET):
        raise io.UnsupportedOperation("seek")

    def tell(self):
        raise io.UnsupportedOperation("tell")


class ExceptHook(QObject):
    # Signal emitted with the `sys.exc_info` tuple.
    handledException = Signal(tuple)

    def __init__(self, parent=None, stream=None, **kwargs):
        super().__init__(parent, **kwargs)
        self.stream = stream

    def __call__(self, exc_type, exc_value, tb):
        # type: (Type[BaseException], BaseException, TracebackType) -> None
        if self.stream is None:
            stream = sys.stderr
        else:
            stream = self.stream
        if stream is not None:
            header = exc_type.__name__ + ' Exception'
            if QThread.currentThread() != QCoreApplication.instance().thread():
                header += " (in non-GUI thread)"
            text = traceback.format_exception(exc_type, exc_value, tb)
            text.insert(0, '{:-^79}\n'.format(' ' + header + ' '))
            text.append('-' * 79 + '\n')
            try:
                stream.writelines(text)
                stream.flush()
            except Exception:
                pass

        self.handledException.emit((exc_type, exc_value, tb))



"""
Tests for WidgetsToolBox.

"""
from typing import cast

from AnyQt.QtWidgets import QWidget, QHBoxLayout
from AnyQt.QtGui import QStandardItemModel
from AnyQt.QtCore import QSize
from AnyQt.QtTest import QTest

from ...registry import tests as registry_tests
from ...registry.qt import QtWidgetRegistry


from ..widgettoolbox import WidgetToolBox, WidgetToolGrid, ToolGrid

from ...gui import test


class TestWidgetToolBox(test.QAppTestCase):
    def setUp(self):
        super().setUp()
        reg = registry_tests.small_testing_registry()
        self.reg = QtWidgetRegistry(reg)

    def tearDown(self):
        self.reg.model().clear()
        del self.reg
        super().tearDown()

    def test_widgettoolgrid(self):
        w = QWidget()
        layout = QHBoxLayout()

        triggered_actions1 = []
        triggered_actions2 = []

        model = self.reg.model()
        data_descriptions = self.reg.widgets("Constants")
        one_action = self.reg.action_for_widget("one")

        actions = list(map(self.reg.action_for_widget, data_descriptions))

        grid = ToolGrid(w)
        grid.setActions(actions)
        grid.actionTriggered.connect(triggered_actions1.append)
        layout.addWidget(grid)

        grid = WidgetToolGrid(w)

        # First category ("Data")
        grid.setModel(model, rootIndex=model.index(0, 0))

        self.assertIs(model, grid.model())

        # Test order of buttons
        grid_layout = grid.layout()
        for i in range(len(actions)):
            button = grid_layout.itemAtPosition(i // 4, i % 4).widget()
            self.assertIs(button.defaultAction(), actions[i])

        grid.actionTriggered.connect(triggered_actions2.append)

        layout.addWidget(grid)

        w.setLayout(layout)
        w.show()
        one_action.trigger()
        self.qWait()

    def test_toolbox(self):
        w = QWidget()
        layout = QHBoxLayout()
        triggered_actions = []

        model = self.reg.model()
        one_action = self.reg.action_for_widget("one")

        box = WidgetToolBox()
        box.setModel(model)
        box.triggered.connect(triggered_actions.append)
        layout.addWidget(box)

        box.setButtonSize(QSize(50, 80))

        w.setLayout(layout)
        w.show()

        one_action.trigger()

        box.setButtonSize(QSize(60, 80))
        box.setIconSize(QSize(35, 35))
        box.setTabButtonHeight(40)
        box.setTabIconSize(QSize(30, 30))

        box.setModel(QStandardItemModel())
        self.assertEqual(box.count(), 0)
        box.setModel(model)
        self.assertEqual(box.count(), model.rowCount())
        self.qWait()

    def test_filter(self):
        w = WidgetToolBox()
        w.setModel(self.reg.model())
        edit = w.filterLineEdit()
        g0 = cast(WidgetToolGrid, w.widget(0))
        self.assertEqual(g0.model().rowCount(g0.rootIndex()), 3)
        QTest.keyClicks(edit, "zero")
        self.assertEqual(g0.model().rowCount(g0.rootIndex()), 1)
        QTest.keyClicks(edit, "\b\b\b\b")
        self.assertEqual(g0.model().rowCount(g0.rootIndex()), 3)



import os
import tempfile
import unittest
from contextlib import contextmanager
from unittest.mock import patch
from zipfile import ZipFile

from AnyQt.QtCore import QEventLoop, QMimeData, QPointF, Qt, QUrl
from AnyQt.QtGui import QDropEvent
from AnyQt.QtTest import QTest
from AnyQt.QtWidgets import QDialogButtonBox, QMessageBox, QTreeView, QStyle
from pkg_resources import Distribution, EntryPoint

from orangecanvas.application import addons
from orangecanvas.application.addons import AddonManagerDialog
from orangecanvas.application.utils.addons import (
    Available,
    CondaInstaller,
    Install,
    Installable,
    Installed,
    PipInstaller,
    Uninstall,
    Upgrade,
    _QueryResult,
)
from orangecanvas.gui.test import QAppTestCase
from orangecanvas.utils.qinvoke import qinvoke


@contextmanager
def addon_archive(pkginfo):
    file = tempfile.NamedTemporaryFile("wb", delete=False, suffix=".zip")
    name = file.name
    file.close()
    with ZipFile(name, 'w') as myzip:
        myzip.writestr('PKG-INFO', pkginfo)
    try:
        yield name
    finally:
        os.remove(name)


class TestAddonManagerDialog(QAppTestCase):
    def test_widget(self):
        items = [
            Installed(
                Installable("foo", "1.1", "", "", "", []),
                Distribution(project_name="foo", version="1.0"),
            ),
            Available(
                Installable("q", "1.2", "", "", "", [])
            ),
            Installed(
                None,
                Distribution(project_name="a", version="0.0")
            ),
        ]
        w = AddonManagerDialog()
        w.setItems(items)
        _ = w.items()
        state = w.itemState()
        self.assertSequenceEqual(state, [])
        state = [(Install, items[1])]
        w.setItemState(state)
        self.assertSequenceEqual(state, w.itemState())
        state = state + [(Upgrade, items[0])]
        w.setItemState(state)
        self.assertSequenceEqual(state, w.itemState()[::-1])
        state = [(Uninstall, items[0])]
        w.setItemState(state)
        self.assertSequenceEqual(state, w.itemState())
        updateTopLayout = w._AddonManagerDialog__updateTopLayout
        updateTopLayout(False)
        updateTopLayout(True)
        w.setItemState([])

        # toggle install state
        view = w.findChild(QTreeView, "add-ons-view")
        index = view.model().index(0, 0)
        delegate = view.itemDelegateForColumn(0)
        style = view.style()
        opt = view.viewOptions()
        opt.rect = view.visualRect(index)
        delegate.initStyleOption(opt, index)

        rect = style.subElementRect(
            QStyle.SE_ItemViewItemCheckIndicator, opt, view
        )

        def check_state_equal(left, right):
            self.assertEqual(Qt.CheckState(left), Qt.CheckState(right))

        check_state_equal(index.data(Qt.CheckStateRole), Qt.PartiallyChecked)
        QTest.mouseClick(view.viewport(), Qt.LeftButton, pos=rect.center())
        check_state_equal(index.data(Qt.CheckStateRole), Qt.Checked)
        QTest.mouseClick(view.viewport(), Qt.LeftButton, pos=rect.center())
        check_state_equal(index.data(Qt.CheckStateRole), Qt.Unchecked)

    @patch("orangecanvas.config.default.addon_entry_points",
           return_value=[EntryPoint(
               "a", "b", dist=Distribution(project_name="foo", version="1.0"))])
    def test_drop(self, p1):
        items = [
            Installed(
                Installable("foo", "1.1", "", "", "", []),
                Distribution(project_name="foo", version="1.0"),
            ),
        ]
        w = AddonManagerDialog()
        w.setItems(items)

        # drop an addon already in the list
        pkginfo = "Metadata-Version: 1.0\nName: foo\nVersion: 0.9"
        with addon_archive(pkginfo) as fn:
            event = self._drop_event(QUrl.fromLocalFile(fn))
            w.dropEvent(event)
        items = w.items()
        self.assertEqual(1, len(items))
        self.assertEqual("0.9", items[0].installable.version)
        self.assertEqual(True, items[0].installable.force)
        state = [(Upgrade, items[0])]
        self.assertSequenceEqual(state, w.itemState())

        # drop a new addon
        pkginfo = "Metadata-Version: 1.0\nName: foo2\nVersion: 0.8"
        with addon_archive(pkginfo) as fn:
            event = self._drop_event(QUrl.fromLocalFile(fn))
            w.dropEvent(event)
        items = w.items()
        self.assertEqual(2, len(items))
        self.assertEqual("0.8", items[1].installable.version)
        self.assertEqual(True, items[1].installable.force)
        state = state + [(Install, items[1])]
        self.assertSequenceEqual(state, w.itemState())

    def _drop_event(self, url):
        # make sure data does not get garbage collected before it used
        # pylint: disable=attribute-defined-outside-init
        self.event_data = data = QMimeData()
        data.setUrls([QUrl(url)])

        return QDropEvent(
            QPointF(0, 0), Qt.MoveAction, data,
            Qt.NoButton, Qt.NoModifier, QDropEvent.Drop)

    def test_run_query(self):
        w = AddonManagerDialog()

        query_res = [
            _QueryResult("uber-pkg", None),
            _QueryResult("unter-pkg", Installable("unter-pkg", "0.0.0", "", "", "", []))
        ]

        def query(names):
            return query_res

        with patch.object(QMessageBox, "exec", return_value=QMessageBox.Cancel), \
             patch.object(addons, "query_pypi", query):
            f = w.runQueryAndAddResults(
                ["uber-pkg", "unter-pkg"],
            )
            loop = QEventLoop()
            f.add_done_callback(qinvoke(lambda f: loop.quit(), loop))
            loop.exec()
            items = w.items()
            self.assertEqual(items, [Available(query_res[1].installable)])

    def test_install(self):
        w = AddonManagerDialog()
        foo = Available(Installable("foo", "1.1", "", "", "", []))
        w.setItems([foo])
        w.setItemState([(Install, foo)])
        with patch.object(PipInstaller, "install", lambda self, pkg: None), \
             patch.object(CondaInstaller, "install", lambda self, pkg: None), \
             patch.object(QMessageBox, "exec", return_value=QMessageBox.Cancel):
            b = w.findChild(QDialogButtonBox)
            b.accepted.emit()
            QTest.qWait(1)
            w.reject()
            QTest.qWait(1)

        w.deleteLater()


if __name__ == "__main__":
    unittest.main()



from itertools import count
from contextlib import contextmanager
from typing import Optional

from AnyQt.QtCore import Qt, QObject, QSize, QRect, QT_VERSION_INFO
from AnyQt.QtGui import (
    QIconEngine, QPalette, QIcon, QPixmap, QPixmapCache, QImage, QPainter
)
from AnyQt.QtWidgets import QApplication, QStyleOption

from orangecanvas.gui.utils import luminance
from orangecanvas.utils.image import grayscale_invert

__all__ = [
    "StyledIconEngine",
    "SymbolIconEngine",
]


_cache_id_gen = count()


class StyledIconEngine(QIconEngine):
    """
    An abstract base class for icon engines that adapt to effective palette.
    """
    __slots__ = ("__palette", "__styleObject")

    def __init__(self, *args, palette: Optional[QPalette] = None,
                 styleObject: Optional[QObject] = None, **kwargs):
        self.__palette = QPalette(palette) if palette is not None else None
        self.__styleObject = styleObject
        super().__init__(*args, **kwargs)

    @staticmethod
    def paletteFromStyleObject(obj: QObject) -> Optional[QPalette]:
        palette = obj.property("palette")
        if isinstance(palette, QPalette):
            return palette
        else:
            return None

    __paletteOverride = None

    @staticmethod
    @contextmanager
    def setOverridePalette(palette: QPalette):
        """
        Temporarily override used QApplication.palette() with this class.

        This can be used when the icon is drawn on a non default background
        and as such might not contrast with it when using the default palette,
        and neither paint device nor styleObject can be used for this.
        """
        old = StyledIconEngine.__paletteOverride
        try:
            StyledIconEngine.__paletteOverride = palette
            yield
        finally:
            StyledIconEngine.__paletteOverride = old

    @staticmethod
    def paletteOverride() -> Optional[QPalette]:
        return StyledIconEngine.__paletteOverride

    def effectivePalette(self) -> QPalette:
        if StyledIconEngine.__paletteOverride is not None:
            return StyledIconEngine.__paletteOverride
        if self.__palette is not None:
            return self.__palette
        elif self.__styleObject is not None:
            palette = self.paletteFromStyleObject(self.__styleObject)
            if palette is not None:
                return palette
        return QApplication.palette()


# shorthands for eliminating runtime attr load in hot path
_QIcon_Active_Modes = (QIcon.Active, QIcon.Selected)
_QIcon_Disabled = QIcon.Disabled

_QPalette_Active = QPalette.Active
_QPalette_WindowText = QPalette.WindowText
_QPalette_Disabled = QPalette.Disabled
_QPalette_HighlightedText = QPalette.HighlightedText


class SymbolIconEngine(StyledIconEngine):
    """
    A *Symbolic* icon engine adapter for turning simple grayscale base icon
    to current effective appearance.

    Arguments
    ---------
    base: QIcon
        The base icon.
    """
    def __init__(self, base: QIcon):
        super().__init__()
        self.__base = QIcon(base)
        self.__cache_key = next(_cache_id_gen)

    def paint(
            self, painter: QPainter, rect: QRect, mode: QIcon.Mode,
            state: QIcon.State
    ) -> None:
        if not self.__base.isNull():
            palette = self.effectivePalette()
            size = rect.size()
            dpr = painter.device().devicePixelRatioF()
            size = size * dpr
            pm = self.__renderStyledPixmap(size, mode, state, palette)
            painter.drawPixmap(rect, pm)

    def pixmap(self, size: QSize, mode: QIcon.Mode, state: QIcon.State) -> QPixmap:
        return self.__renderStyledPixmap(size, mode, state, self.effectivePalette())

    def __renderStyledPixmap(
            self, size: QSize, mode: QIcon.Mode, state: QIcon.State,
            palette: QPalette
    ) -> QPixmap:
        active = mode in _QIcon_Active_Modes
        disabled = mode == _QIcon_Disabled
        cg = _QPalette_Disabled if disabled else _QPalette_Active
        role = _QPalette_WindowText if active else _QPalette_HighlightedText
        namespace = f"{__name__}:SymbolIconEngine/{self.__cache_key}"
        cachekey = f"{size.width()}x{size.height()}"
        style_key = f"{hex(palette.cacheKey())}-{cg}-{role}"
        pmcachekey = f"{namespace}/{cachekey}/{style_key}"
        pm = QPixmapCache.find(pmcachekey)
        if pm is None or pm.isNull():
            color = palette.color(QPalette.Text)
            src = qicon_pixmap(self.__base, size, 1.0, mode, state)
            src = src.toImage().convertToFormat(QImage.Format_ARGB32_Premultiplied)
            if luminance(color) > 0.5:
                dest = grayscale_invert(
                    src,
                    palette.color(QPalette.Text),
                    palette.color(QPalette.Base),
                )
            else:
                dest = src
            pm = QPixmap.fromImage(dest)
            QPixmapCache.insert(pmcachekey, pm)

        self.__style = style = QApplication.style()
        if style is not None:
            opt = QStyleOption()
            opt.palette = palette
            pm = style.generatedIconPixmap(mode, pm, opt)
        return pm

    def clone(self) -> 'QIconEngine':
        return SymbolIconEngine(self.__base)


def qicon_pixmap(
        base: QIcon, size: QSize, scale: float, mode: QIcon.Mode,
        state: QIcon.State
) -> QPixmap:
    """
    Like QIcon.pixmap(size: QSize, scale: float, ...) overload in Qt6.

    On Qt 6 this directly calls the corresponding overload.
    On Qt 5 this is emulated by painting on a suitable constructed pixmap.
    """
    size = base.actualSize(size * scale, mode, state)
    pm = QPixmap(size)
    pm.setDevicePixelRatio(scale)
    pm.fill(Qt.transparent)
    p = QPainter(pm)
    base.paint(p, 0, 0, size.width(), size.height(), Qt.AlignCenter, mode, state)
    p.end()
    return pm


if QT_VERSION_INFO >= (6, 0):
    qicon_pixmap = QIcon.pixmap



import os
import stat
import unittest
from tempfile import mkdtemp

from pkg_resources import Requirement
from requests import Session
from requests_cache import CachedSession

from orangecanvas.application.utils.addons import (
    Available,
    Installable,
    Installed,
    installable_from_json_response,
    installable_items,
    is_updatable,
    prettify_name, _session,
)
from orangecanvas.config import Distribution


class TestUtils(unittest.TestCase):
    def test_items_1(self):
        inst = Installable("foo", "1.0", "a foo", "", "", [])
        dist = Distribution(project_name="foo", version="1.0")
        item = Available(inst)
        self.assertFalse(is_updatable(item))

        item = Installed(None, dist)
        self.assertFalse(is_updatable(item))
        item = Installed(inst, dist)
        self.assertFalse(is_updatable(item))

        item = Installed(inst._replace(version="0.9"), dist)
        self.assertFalse(is_updatable(item))

        item = Installed(inst._replace(version="1.1"), dist)
        self.assertTrue(is_updatable(item))

        item = Installed(inst._replace(version="2.0"), dist,
                         constraint=Requirement.parse("foo<1.99"))
        self.assertFalse(is_updatable(item))
        item = Installed(inst._replace(version="2.0"), dist,
                         constraint=Requirement.parse("foo<2.99"))
        self.assertTrue(is_updatable(item))

    def test_items_2(self):
        inst1 = Installable("foo", "1.0", "a foo", "", "", [])
        inst2 = Installable("bar", "1.0", "a bar", "", "", [])
        dist2 = Distribution(project_name="bar", version="0.9")
        dist3 = Distribution(project_name="quack", version="1.0")
        items = installable_items([inst1, inst2], [dist2, dist3])
        self.assertIn(Available(inst1), items)
        self.assertIn(Installed(inst2, dist2), items)
        self.assertIn(Installed(None, dist3), items)

    def test_installable_from_json_response(self):
        inst = installable_from_json_response({
            "info": {
                "name": "foo",
                "version": "1.0",
            },
            "releases": {
                "1.0": [
                    {
                        "filename": "aa.tar.gz",
                        "url": "https://examples.com",
                        "size": 100,
                        "packagetype": "sdist",
                    }
                ]
            },
        })
        self.assertTrue(inst.name, "foo")
        self.assertEqual(inst.version, "1.0")

    def test_prettify_name(self):
        names = [
            'AFooBar', 'FooBar', 'Foo-Bar', 'Foo-Bar-FOOBAR',
            'Foo-bar-foobar', 'Foo', 'FOOBar', 'A4FooBar',
            '4Foo', 'Foo3Bar'
        ]
        pretty_names = [
            'A Foo Bar', 'Foo Bar', 'Foo Bar', 'Foo Bar FOOBAR',
            'Foo bar foobar', 'Foo', 'FOO Bar', 'A4Foo Bar',
            '4Foo', 'Foo3Bar'
        ]

        for name, pretty_name in zip(names, pretty_names):
            self.assertEqual(pretty_name, prettify_name(name))

        # test if orange prefix is handled
        self.assertEqual('Orange', prettify_name('Orange'))
        self.assertEqual('Orange3', prettify_name('Orange3'))
        self.assertEqual('Some Addon', prettify_name('Orange-SomeAddon'))
        self.assertEqual('Text', prettify_name('Orange3-Text'))
        self.assertEqual('Image Analytics', prettify_name('Orange3-ImageAnalytics'))
        self.assertEqual('Survival Analysis', prettify_name('Orange3-Survival-Analysis'))

    def test_session(self):
        # when permissions - use CachedSession
        self.assertIsInstance(_session(), CachedSession)

        # when no permissions - use request's Session
        temp_dir = mkdtemp()
        os.chmod(temp_dir, stat.S_IRUSR)
        self.assertIsInstance(_session(temp_dir), Session)


if __name__ == "__main__":
    unittest.main()



import itertools
from collections import deque
from typing import TypeVar, Iterable, Callable, Set, List, Hashable

__all__ = [
    "traverse_bf", "strongly_connected_components"
]

H = TypeVar("H", bound=Hashable)


def traverse_bf(start, expand):
    # type: (H, Callable[[H], Iterable[H]]) -> Iterable[H]
    """
    Breadth first traversal of a DAG starting from `start`.

    Parameters
    ----------
    start : H
        A starting node
    expand : (H) -> Iterable[H]
        A function returning children of a node.
    """
    queue = deque([start])
    visited = set()  # type: Set[H]
    while queue:
        item = queue.popleft()
        if item not in visited:
            yield item
            visited.add(item)
            queue.extend(expand(item))


def strongly_connected_components(nodes, expand):
    # type: (Iterable[H], Callable[[H], Iterable[H]]) -> List[List[H]]
    """
    Return a list of strongly connected components.

    Implementation of Tarjan's SCC algorithm.
    """
    # SCC found
    components = []  # type: List[List[H]]
    # node stack in BFS
    stack = []       # type: List[H]
    # == set(stack) : a set of all nodes in stack (for faster lookup)
    stackset = set()

    # node -> int increasing node numbering as encountered in DFS traversal
    index = {}
    # node -> int the lowest node index reachable from a node
    lowlink = {}

    indexgen = itertools.count()

    def push_node(v):
        # type: (H) -> None
        """Push node onto the stack."""
        stack.append(v)
        stackset.add(v)
        index[v] = lowlink[v] = next(indexgen)

    def pop_scc(v):
        # type: (H) -> List[H]
        """Pop from the stack a SCC rooted at node v."""
        i = stack.index(v)
        scc = stack[i:]
        del stack[i:]
        stackset.difference_update(scc)
        return scc

    def isvisited(node):  # type: (H) -> bool
        return node in index

    def strong_connect(v):
        # type: (H) -> None
        push_node(v)

        for w in expand(v):
            if not isvisited(w):
                strong_connect(w)
                lowlink[v] = min(lowlink[v], lowlink[w])
            elif w in stackset:
                lowlink[v] = min(lowlink[v], index[w])

        if index[v] == lowlink[v]:
            scc = pop_scc(v)
            components.append(scc)

    for node in nodes:
        if not isvisited(node):
            strong_connect(node)

    return components



"""
"""

from ...gui import test
from ...registry.tests import small_testing_registry
from ...registry import InputSignal, OutputSignal

from .. import SchemeNode


class TestScheme(test.QAppTestCase):
    def test_node(self):
        """Test SchemeNode.
        """
        reg = small_testing_registry()
        one_desc = reg.widget("one")

        node = SchemeNode(one_desc)

        inputs = node.input_channels()
        self.assertSequenceEqual(inputs, one_desc.inputs)
        for ch in inputs:
            channel = node.input_channel(ch.name)
            self.assertIsInstance(channel, InputSignal)
            self.assertTrue(channel in inputs)
        self.assertRaises(ValueError, node.input_channel, "%%&&&$$()[()[")

        outputs = node.output_channels()
        self.assertSequenceEqual(outputs, one_desc.outputs)
        for ch in outputs:
            channel = node.output_channel(ch.name)
            self.assertIsInstance(channel, OutputSignal)
            self.assertTrue(channel in outputs)
        self.assertRaises(ValueError, node.output_channel, "%%&&&$$()[()[")

    def test_channels_by_name_or_id(self):
        reg = small_testing_registry()

        zero_desc = reg.widget("zero")
        node = SchemeNode(zero_desc)
        self.assertIs(node.output_channel("value"), zero_desc.outputs[0])
        self.assertIs(node.output_channel("val"), zero_desc.outputs[0])

        add_desc = reg.widget("add")
        node = SchemeNode(add_desc)
        self.assertIs(node.input_channel("left"), add_desc.inputs[0])
        self.assertIs(node.input_channel("right"), add_desc.inputs[1])
        self.assertIs(node.input_channel("droite"), add_desc.inputs[1])
        self.assertRaises(ValueError, node.input_channel, "gauche")



"""
=========================
User Interaction Handlers
=========================

User interaction handlers for a :class:`~.SchemeEditWidget`.

User interactions encapsulate the logic of user interactions with the
scheme document.

All interactions are subclasses of :class:`UserInteraction`.


"""
import typing
from typing import Optional, Any, Tuple, List, Set, Iterable, Sequence, Dict

import abc
import logging
from functools import reduce

from AnyQt.QtWidgets import (
    QApplication, QGraphicsRectItem, QGraphicsSceneMouseEvent,
    QGraphicsSceneContextMenuEvent, QWidget, QGraphicsItem,
    QGraphicsSceneDragDropEvent, QMenu, QAction
)
from AnyQt.QtGui import QPen, QBrush, QColor, QFontMetrics, QKeyEvent, QFont
from AnyQt.QtCore import (
    Qt, QObject, QCoreApplication, QSizeF, QPointF, QRect, QRectF, QLineF,
    QPoint, QMimeData,
)
from AnyQt.QtCore import pyqtSignal as Signal

from orangecanvas.document.commands import UndoCommand
from .usagestatistics import UsageStatistics
from ..registry.description import WidgetDescription, OutputSignal, InputSignal
from ..registry.qt import QtWidgetRegistry, tooltip_helper, whats_this_helper
from .. import scheme
from ..scheme import (
    SchemeNode as Node, SchemeLink as Link, Scheme, WorkflowEvent,
    compatible_channels
)
from ..canvas import items
from ..canvas.items import controlpoints
from ..gui.quickhelp import QuickHelpTipEvent
from . import commands
from .editlinksdialog import EditLinksDialog
from ..utils import unique

if typing.TYPE_CHECKING:
    from .schemeedit import SchemeEditWidget

    A = typing.TypeVar("A")
    #: Output/Input pair of a link
    OIPair = Tuple[OutputSignal, InputSignal]

try:
    from importlib.metadata import EntryPoint, entry_points
except ImportError:
    from importlib_metadata import EntryPoint, entry_points


log = logging.getLogger(__name__)


def assert_not_none(optional):
    # type: (Optional[A]) -> A
    assert optional is not None
    return optional


class UserInteraction(QObject):
    """
    Base class for user interaction handlers.

    Parameters
    ----------
    document : :class:`~.SchemeEditWidget`
        An scheme editor instance with which the user is interacting.
    parent : :class:`QObject`, optional
        A parent QObject
    deleteOnEnd : bool, optional
        Should the UserInteraction be deleted when it finishes (``True``
        by default).

    """
    # Cancel reason flags

    #: No specified reason
    NoReason = 0
    #: User canceled the operation (e.g. pressing ESC)
    UserCancelReason = 1
    #: Another interaction was set
    InteractionOverrideReason = 3
    #: An internal error occurred
    ErrorReason = 4
    #: Other (unspecified) reason
    OtherReason = 5

    #: Emitted when the interaction is set on the scene.
    started = Signal()

    #: Emitted when the interaction finishes successfully.
    finished = Signal()

    #: Emitted when the interaction ends (canceled or finished)
    ended = Signal()

    #: Emitted when the interaction is canceled.
    canceled = Signal([], [int])

    def __init__(self, document, parent=None, deleteOnEnd=True):
        # type: ('SchemeEditWidget', Optional[QObject], bool) -> None
        super().__init__(parent)
        self.document = document
        self.scene = document.scene()
        scheme_ = document.scheme()
        assert scheme_ is not None
        self.scheme = scheme_  # type: scheme.Scheme
        self.suggestions = document.suggestions()
        self.deleteOnEnd = deleteOnEnd

        self.cancelOnEsc = False

        self.__finished = False
        self.__canceled = False
        self.__cancelReason = self.NoReason

    def start(self):
        # type: () -> None
        """
        Start the interaction. This is called by the :class:`CanvasScene` when
        the interaction is installed.

        .. note:: Must be called from subclass implementations.

        """
        self.started.emit()

    def end(self):
        # type: () -> None
        """
        Finish the interaction. Restore any leftover state in this method.

        .. note:: This gets called from the default :func:`cancel`
                  implementation.

        """
        self.__finished = True

        if self.scene.user_interaction_handler is self:
            self.scene.set_user_interaction_handler(None)

        if self.__canceled:
            self.canceled.emit()
            self.canceled[int].emit(self.__cancelReason)
        else:
            self.finished.emit()

        self.ended.emit()

        if self.deleteOnEnd:
            self.deleteLater()

    def cancel(self, reason=OtherReason):
        # type: (int) -> None
        """
        Cancel the interaction with `reason`.
        """

        self.__canceled = True
        self.__cancelReason = reason

        self.end()

    def isFinished(self):
        # type: () -> bool
        """
        Is the interaction finished.
        """
        return self.__finished

    def isCanceled(self):
        # type: () -> bool
        """
        Was the interaction canceled.
        """
        return self.__canceled

    def cancelReason(self):
        # type: () -> int
        """
        Return the reason the interaction was canceled.
        """
        return self.__cancelReason

    def mousePressEvent(self, event):
        # type: (QGraphicsSceneMouseEvent) -> bool
        """
        Handle a `QGraphicsScene.mousePressEvent`.
        """
        return False

    def mouseMoveEvent(self, event):
        # type: (QGraphicsSceneMouseEvent) -> bool
        """
        Handle a `GraphicsScene.mouseMoveEvent`.
        """
        return False

    def mouseReleaseEvent(self, event):
        # type: (QGraphicsSceneMouseEvent) -> bool
        """
        Handle a `QGraphicsScene.mouseReleaseEvent`.
        """
        return False

    def mouseDoubleClickEvent(self, event):
        # type: (QGraphicsSceneMouseEvent) -> bool
        """
        Handle a `QGraphicsScene.mouseDoubleClickEvent`.
        """
        return False

    def keyPressEvent(self, event):
        # type: (QKeyEvent) -> bool
        """
        Handle a `QGraphicsScene.keyPressEvent`
        """
        if self.cancelOnEsc and event.key() == Qt.Key_Escape:
            self.cancel(self.UserCancelReason)
        return False

    def keyReleaseEvent(self, event):
        # type: (QKeyEvent) -> bool
        """
        Handle a `QGraphicsScene.keyPressEvent`
        """
        return False

    def contextMenuEvent(self, event):
        # type: (QGraphicsSceneContextMenuEvent) -> bool
        """
        Handle a `QGraphicsScene.contextMenuEvent`
        """
        return False

    def dragEnterEvent(self, event):
        # type: (QGraphicsSceneDragDropEvent) -> bool
        """
        Handle a `QGraphicsScene.dragEnterEvent`

        .. versionadded:: 0.1.20
        """
        return False

    def dragMoveEvent(self, event):
        # type: (QGraphicsSceneDragDropEvent) -> bool
        """
        Handle a `QGraphicsScene.dragMoveEvent`

        .. versionadded:: 0.1.20
        """
        return False

    def dragLeaveEvent(self, event):
        # type: (QGraphicsSceneDragDropEvent) -> bool
        """
        Handle a `QGraphicsScene.dragLeaveEvent`

        .. versionadded:: 0.1.20
        """
        return False

    def dropEvent(self, event):
        # type: (QGraphicsSceneDragDropEvent) -> bool
        """
        Handle a `QGraphicsScene.dropEvent`

        .. versionadded:: 0.1.20
        """
        return False


class NoPossibleLinksError(ValueError):
    pass


class UserCanceledError(ValueError):
    pass


def reversed_arguments(func):
    """
    Return a function with reversed argument order.
    """
    def wrapped(*args):
        return func(*reversed(args))
    return wrapped


class NewLinkAction(UserInteraction):
    """
    User drags a new link from an existing `NodeAnchorItem` to create
    a connection between two existing nodes or to a new node if the release
    is over an empty area, in which case a quick menu for new node selection
    is presented to the user.

    """
    # direction of the drag
    FROM_SOURCE = 1
    FROM_SINK = 2

    def __init__(self, document, *args, **kwargs):
        super().__init__(document, *args, **kwargs)
        self.from_item = None    # type: Optional[items.NodeItem]
        self.from_signal = None  # type: Optional[Union[InputSignal, OutputSignal]]
        self.direction = 0       # type: int
        self.showing_incompatible_widget = False  # type: bool

        # An `NodeItem` currently under the mouse as a possible
        # link drop target.
        self.current_target_item = None  # type: Optional[items.NodeItem]
        # A temporary `LinkItem` used while dragging.
        self.tmp_link_item = None        # type: Optional[items.LinkItem]
        # An temporary `AnchorPoint` inserted into `current_target_item`
        self.tmp_anchor_point = None     # type: Optional[items.AnchorPoint]
        # An `AnchorPoint` following the mouse cursor
        self.cursor_anchor_point = None  # type: Optional[items.AnchorPoint]
        # An UndoCommand
        self.macro = None  # type: Optional[UndoCommand]

        # Cache viable signals of currently hovered node
        self.__target_compatible_signals: Sequence[Tuple[OutputSignal, InputSignal]] = []

        self.cancelOnEsc = True

    def remove_tmp_anchor(self):
        # type: () -> None
        """
        Remove a temporary anchor point from the current target item.
        """
        assert self.current_target_item is not None
        assert self.tmp_anchor_point is not None
        if self.direction == self.FROM_SOURCE:
            self.current_target_item.removeInputAnchor(self.tmp_anchor_point)
        else:
            self.current_target_item.removeOutputAnchor(self.tmp_anchor_point)
        self.tmp_anchor_point = None

    def update_tmp_anchor(self, item, scenePos):
        # type: (items.NodeItem, QPointF) -> None
        """
        If hovering over a new compatible channel, move it.
        """
        assert self.tmp_anchor_point is not None
        if self.direction == self.FROM_SOURCE:
            signal = item.inputAnchorItem.signalAtPos(scenePos,
                                                      self.__target_compatible_signals)
        else:
            signal = item.outputAnchorItem.signalAtPos(scenePos,
                                                       self.__target_compatible_signals)
        self.tmp_anchor_point.setSignal(signal)

    def create_tmp_anchor(self, item, scenePos):
        # type: (items.NodeItem, QPointF) -> None
        """
        Create a new tmp anchor at the `item` (:class:`NodeItem`).
        """
        assert self.tmp_anchor_point is None
        if self.direction == self.FROM_SOURCE:
            anchor = item.inputAnchorItem
            signal = anchor.signalAtPos(scenePos,
                                        self.__target_compatible_signals)
            self.tmp_anchor_point = item.newInputAnchor(signal)
        else:
            anchor = item.outputAnchorItem
            signal = anchor.signalAtPos(scenePos,
                                        self.__target_compatible_signals)
            self.tmp_anchor_point = item.newOutputAnchor(signal)

    def __possible_connection_signal_pairs(
            self, target_item: items.NodeItem
    ) -> Sequence[Tuple[OutputSignal, InputSignal]]:
        """
        Return possible connection signal pairs between current
        `self.from_item` and `target_item`.
        """
        if self.from_item is None:
            return []
        node1 = self.scene.node_for_item(self.from_item)
        node2 = self.scene.node_for_item(target_item)

        if self.direction == self.FROM_SOURCE:
            links = self.scheme.propose_links(node1, node2,
                                              source_signal=self.from_signal)
        else:
            links = self.scheme.propose_links(node2, node1,
                                              sink_signal=self.from_signal)
        return [(s1, s2) for s1, s2, _ in links]

    def can_connect(self, target_item):
        # type: (items.NodeItem) -> bool
        """
        Is the connection between `self.from_item` (item where the drag
        started) and `target_item` possible.
        """
        return bool(self.__possible_connection_signal_pairs(target_item))

    def set_link_target_anchor(self, anchor):
        # type: (items.AnchorPoint) -> None
        """
        Set the temp line target anchor.
        """
        assert self.tmp_link_item is not None
        if self.direction == self.FROM_SOURCE:
            self.tmp_link_item.setSinkItem(None, anchor=anchor)
        else:
            self.tmp_link_item.setSourceItem(None, anchor=anchor)

    def target_node_item_at(self, pos):
        # type: (QPointF) -> Optional[items.NodeItem]
        """
        Return a suitable :class:`NodeItem` at position `pos` on which
        a link can be dropped.
        """
        # Test for a suitable `NodeAnchorItem` or `NodeItem` at pos.
        if self.direction == self.FROM_SOURCE:
            anchor_type = items.SinkAnchorItem
        else:
            anchor_type = items.SourceAnchorItem

        item = self.scene.item_at(pos, (anchor_type, items.NodeItem))

        if isinstance(item, anchor_type):
            return item.parentNodeItem()
        elif isinstance(item, items.NodeItem):
            return item
        else:
            return None

    def mousePressEvent(self, event):
        # type: (QGraphicsSceneMouseEvent) -> bool
        anchor_item = self.scene.item_at(
            event.scenePos(), items.NodeAnchorItem
        )
        if anchor_item is not None and event.button() == Qt.LeftButton:
            # Start a new link starting at item
            self.from_item = anchor_item.parentNodeItem()
            if isinstance(anchor_item, items.SourceAnchorItem):
                self.direction = NewLinkAction.FROM_SOURCE
            else:
                self.direction = NewLinkAction.FROM_SINK

            event.accept()

            helpevent = QuickHelpTipEvent(
                self.tr("Create a new link"),
                self.tr('<h3>Create new link</h3>'
                        '<p>Drag a link to an existing node or release on '
                        'an empty spot to create a new node.</p>'
                        '<p>Hold Shift when releasing the mouse button to '
                        'edit connections.</p>'
#                        '<a href="help://orange-canvas/create-new-links">'
#                        'More ...</a>'
                        )
            )
            QCoreApplication.postEvent(self.document, helpevent)
            return True
        else:
            # Whoever put us in charge did not know what he was doing.
            self.cancel(self.ErrorReason)
            return False

    def mouseMoveEvent(self, event):
        # type: (QGraphicsSceneMouseEvent) -> bool
        if self.tmp_link_item is None:
            # On first mouse move event create the temp link item and
            # initialize it to follow the `cursor_anchor_point`.
            self.tmp_link_item = items.LinkItem()
            # An anchor under the cursor for the duration of this action.
            self.cursor_anchor_point = items.AnchorPoint()
            self.cursor_anchor_point.setPos(event.scenePos())

            # Set the `fixed` end of the temp link (where the drag started).
            scenePos = event.scenePos()

            if self.direction == self.FROM_SOURCE:
                anchor = self.from_item.outputAnchorItem
            else:
                anchor = self.from_item.inputAnchorItem
            anchor.setHovered(False)
            anchor.setCompatibleSignals(None)

            if anchor.anchorOpen():
                signal = anchor.signalAtPos(scenePos)
                anchor.setKeepAnchorOpen(signal)
            else:
                signal = None
            self.from_signal = signal

            if self.direction == self.FROM_SOURCE:
                self.tmp_link_item.setSourceItem(self.from_item, signal)
            else:
                self.tmp_link_item.setSinkItem(self.from_item, signal)

            self.set_link_target_anchor(self.cursor_anchor_point)
            self.scene.addItem(self.tmp_link_item)

        assert self.cursor_anchor_point is not None

        # `NodeItem` at the cursor position
        item = self.target_node_item_at(event.scenePos())

        if self.current_target_item is not None and \
                (item is None or item is not self.current_target_item):
            # `current_target_item` is no longer under the mouse cursor
            # (was replaced by another item or the the cursor was moved over
            # an empty scene spot.
            log.info("%r is no longer the target.", self.current_target_item)
            if self.direction == self.FROM_SOURCE:
                anchor = self.current_target_item.inputAnchorItem
            else:
                anchor = self.current_target_item.outputAnchorItem
            if self.showing_incompatible_widget:
                anchor.setIncompatible(False)
                self.showing_incompatible_widget = False
            else:
                self.remove_tmp_anchor()
            anchor.setHovered(False)
            anchor.setCompatibleSignals(None)
            self.current_target_item = None

        if item is not None and item is not self.from_item:
            # The mouse is over a node item (different from the starting node)
            if self.current_target_item is item:
                # Mouse is over the same item
                scenePos = event.scenePos()
                # Move to new potential anchor
                if not self.showing_incompatible_widget:
                    self.update_tmp_anchor(item, scenePos)
                else:
                    self.set_link_target_anchor(self.cursor_anchor_point)
            elif self.can_connect(item):
                # Mouse is over a new item
                links = self.__possible_connection_signal_pairs(item)
                log.info("%r is the new target.", item)
                if self.direction == self.FROM_SOURCE:
                    self.__target_compatible_signals = [s2 for s1, s2 in links]
                    item.inputAnchorItem.setCompatibleSignals(
                        self.__target_compatible_signals)
                    item.inputAnchorItem.setHovered(True)
                else:
                    self.__target_compatible_signals = [s1 for s1, s2 in links]
                    item.outputAnchorItem.setCompatibleSignals(
                        self.__target_compatible_signals)
                    item.outputAnchorItem.setHovered(True)
                scenePos = event.scenePos()
                self.create_tmp_anchor(item, scenePos)
                self.set_link_target_anchor(
                    assert_not_none(self.tmp_anchor_point)
                )
                self.current_target_item = item
                self.showing_incompatible_widget = False
            else:
                log.info("%r does not have compatible channels", item)
                self.__target_compatible_signals = []
                if self.direction == self.FROM_SOURCE:
                    anchor = item.inputAnchorItem
                else:
                    anchor = item.outputAnchorItem
                anchor.setCompatibleSignals(
                    self.__target_compatible_signals)
                anchor.setHovered(True)
                anchor.setIncompatible(True)
                self.showing_incompatible_widget = True
                self.set_link_target_anchor(self.cursor_anchor_point)
                self.current_target_item = item
        else:
            self.showing_incompatible_widget = item is not None
            self.__target_compatible_signals = []
            self.set_link_target_anchor(self.cursor_anchor_point)

        self.cursor_anchor_point.setPos(event.scenePos())

        return True

    def mouseReleaseEvent(self, event):
        # type: (QGraphicsSceneMouseEvent) -> bool
        if self.tmp_link_item is not None:
            item = self.target_node_item_at(event.scenePos())
            node = None  # type: Optional[Node]
            stack = self.document.undoStack()

            self.macro = UndoCommand(self.tr("Add link"))

            if item:
                # If the release was over a node item then connect them
                node = self.scene.node_for_item(item)
            else:
                # Release on an empty canvas part
                # Show a quick menu popup for a new widget creation.
                try:
                    node = self.create_new(event)
                except Exception:
                    log.error("Failed to create a new node, ending.",
                              exc_info=True)
                    node = None

                if node is not None:
                    commands.AddNodeCommand(self.scheme, node,
                                            parent=self.macro)

            if node is not None and not self.showing_incompatible_widget:
                if self.direction == self.FROM_SOURCE:
                    source_node = self.scene.node_for_item(self.from_item)
                    source_signal = self.from_signal
                    sink_node = node
                    if item is not None and item.inputAnchorItem.anchorOpen():
                        sink_signal = item.inputAnchorItem.signalAtPos(
                            event.scenePos(),
                            self.__target_compatible_signals
                        )
                    else:
                        sink_signal = None
                else:
                    source_node = node
                    if item is not None and item.outputAnchorItem.anchorOpen():
                        source_signal = item.outputAnchorItem.signalAtPos(
                            event.scenePos(),
                            self.__target_compatible_signals
                        )
                    else:
                        source_signal = None
                    sink_node = self.scene.node_for_item(self.from_item)
                    sink_signal = self.from_signal
                self.suggestions.set_direction(self.direction)
                self.connect_nodes(source_node, sink_node,
                                   source_signal, sink_signal)
                self.reset_open_anchor()
                if not self.isCanceled() or not self.isFinished() and \
                        self.macro is not None:
                    # Push (commit) the add link/node action on the stack.
                    stack.push(self.macro)

            self.end()
            return True
        else:
            self.end()
            return False

    def create_new(self, event):
        # type: (QGraphicsSceneMouseEvent) -> Optional[Node]
        """
        Create and return a new node with a `QuickMenu`.
        """
        pos = event.screenPos()
        menu = self.document.quickMenu()
        node = self.scene.node_for_item(self.from_item)
        from_signal = self.from_signal
        from_desc = node.description

        def is_compatible(
                source_signal: OutputSignal,
                source: WidgetDescription,
                sink: WidgetDescription,
                sink_signal: InputSignal
        ) -> bool:
            return any(scheme.compatible_channels(output, input)
                       for output
                       in ([source_signal] if source_signal else source.outputs)
                       for input
                       in ([sink_signal] if sink_signal else sink.inputs))

        from_sink = self.direction == self.FROM_SINK
        if from_sink:
            # Reverse the argument order.
            is_compatible = reversed_arguments(is_compatible)
            suggestion_sort = self.suggestions.get_source_suggestions(from_desc.name)
        else:
            suggestion_sort = self.suggestions.get_sink_suggestions(from_desc.name)

        def sort(left, right):
            # list stores frequencies, so sign is flipped
            return suggestion_sort[left] > suggestion_sort[right]

        menu.setSortingFunc(sort)

        def filter(index):
            desc = index.data(QtWidgetRegistry.WIDGET_DESC_ROLE)
            if isinstance(desc, WidgetDescription):
                return is_compatible(from_signal, from_desc, desc, None)
            else:
                return False

        menu.setFilterFunc(filter)
        menu.triggerSearch()
        try:
            action = menu.exec(pos)
        finally:
            menu.setFilterFunc(None)

        if action:
            item = action.property("item")
            desc = item.data(QtWidgetRegistry.WIDGET_DESC_ROLE)
            pos = event.scenePos()
            # a new widget should be placed so that the connection
            # stays as it was
            offset = 31 * (-1 if self.direction == self.FROM_SINK else
                           1 if self.direction == self.FROM_SOURCE else 0)
            statistics = self.document.usageStatistics()
            statistics.begin_extend_action(from_sink, node)
            node = self.document.newNodeHelper(desc,
                                               position=(pos.x() + offset,
                                                         pos.y()))
            return node
        else:
            return None

    def connect_nodes(
            self, source_node: Node, sink_node: Node,
            source_signal: Optional[OutputSignal] = None,
            sink_signal: Optional[InputSignal] = None
    ) -> None:
        """
        Connect `source_node` to `sink_node`. If there are more then one
        equally weighted and non conflicting links possible present a
        detailed dialog for link editing.

        """
        UsageStatistics.set_sink_anchor_open(sink_signal is not None)
        UsageStatistics.set_source_anchor_open(source_signal is not None)
        try:
            possible = self.scheme.propose_links(source_node, sink_node,
                                                 source_signal, sink_signal)

            log.debug("proposed (weighted) links: %r",
                      [(s1.name, s2.name, w) for s1, s2, w in possible])

            if not possible:
                raise NoPossibleLinksError

            source, sink, w = possible[0]

            # just a list of signal tuples for now, will be converted
            # to SchemeLinks later
            links_to_add = []     # type: List[Link]
            links_to_remove = []  # type: List[Link]
            show_link_dialog = False

            # Ambiguous new link request.
            if len(possible) >= 2:
                # Check for possible ties in the proposed link weights
                _, _, w2 = possible[1]
                if w == w2:
                    show_link_dialog = True

                # Check for destructive action (i.e. would the new link
                # replace a previous link) except for explicit only link
                # candidates
                if sink.single and w2 > 0 and \
                        self.scheme.find_links(sink_node=sink_node,
                                               sink_channel=sink):
                    show_link_dialog = True

            if show_link_dialog:
                existing = self.scheme.find_links(source_node=source_node,
                                                  sink_node=sink_node)

                if existing:
                    # edit_links will populate the view with existing links
                    initial_links = None
                else:
                    initial_links = [(source, sink)]

                try:
                    rstatus, links_to_add, links_to_remove = self.edit_links(
                        source_node, sink_node, initial_links
                    )
                except Exception:
                    log.error("Failed to edit the links",
                              exc_info=True)
                    raise
                if rstatus == EditLinksDialog.Rejected:
                    raise UserCanceledError
            else:
                # links_to_add now needs to be a list of actual SchemeLinks
                links_to_add = [
                    scheme.SchemeLink(source_node, source, sink_node, sink)
                ]
                links_to_add, links_to_remove = \
                    add_links_plan(self.scheme, links_to_add)

            # Remove temp items before creating any new links
            self.cleanup()

            for link in links_to_remove:
                commands.RemoveLinkCommand(self.scheme, link,
                                           parent=self.macro)

            for link in links_to_add:
                # Check if the new requested link is a duplicate of an
                # existing link
                duplicate = self.scheme.find_links(
                    link.source_node, link.source_channel,
                    link.sink_node, link.sink_channel
                )

                if not duplicate:
                    commands.AddLinkCommand(self.scheme, link,
                                            parent=self.macro)

        except scheme.IncompatibleChannelTypeError:
            log.info("Cannot connect: invalid channel types.")
            self.cancel()
        except scheme.SchemeTopologyError:
            log.info("Cannot connect: connection creates a cycle.")
            self.cancel()
        except NoPossibleLinksError:
            log.info("Cannot connect: no possible links.")
            self.cancel()
        except UserCanceledError:
            log.info("User canceled a new link action.")
            self.cancel(UserInteraction.UserCancelReason)
        except Exception:
            log.error("An error occurred during the creation of a new link.",
                      exc_info=True)
            self.cancel()

    def edit_links(
            self,
            source_node: Node,
            sink_node: Node,
            initial_links: 'Optional[List[OIPair]]' = None
    ) -> 'Tuple[int, List[Link], List[Link]]':
        """
        Show and execute the `EditLinksDialog`.
        Optional `initial_links` list can provide a list of initial
        `(source, sink)` channel tuples to show in the view, otherwise
        the dialog is populated with existing links in the scheme (passing
        an empty list will disable all initial links).

        """
        status, links_to_add_spec, links_to_remove_spec = \
            edit_links(
                self.scheme, source_node, sink_node, initial_links,
                parent=self.document
            )

        if status == EditLinksDialog.Accepted:
            links_to_add = [
                scheme.SchemeLink(
                    source_node, source_channel,
                    sink_node, sink_channel
                ) for source_channel, sink_channel in links_to_add_spec
            ]
            links_to_remove = list(reduce(
                list.__iadd__, (
                    self.scheme.find_links(
                        source_node, source_channel,
                        sink_node, sink_channel
                    ) for source_channel, sink_channel in links_to_remove_spec
                ),
                []
            ))  # type: List[Link]
            conflicting = [conflicting_single_link(self.scheme, link)
                           for link in links_to_add]
            conflicting = [link for link in conflicting if link is not None]
            for link in conflicting:
                if link not in links_to_remove:
                    links_to_remove.append(link)

            return status, links_to_add, links_to_remove
        else:
            return status, [], []

    def end(self):
        # type: () -> None
        self.cleanup()
        self.reset_open_anchor()
        # Remove the help tip set in mousePressEvent
        self.macro = None
        helpevent = QuickHelpTipEvent("", "")
        QCoreApplication.postEvent(self.document, helpevent)
        super().end()

    def cancel(self, reason=UserInteraction.OtherReason):
        # type: (int) -> None
        self.cleanup()
        self.reset_open_anchor()
        super().cancel(reason)

    def cleanup(self):
        # type: () -> None
        """
        Cleanup all temporary items in the scene that are left.
        """
        if self.tmp_link_item:
            self.tmp_link_item.setSinkItem(None)
            self.tmp_link_item.setSourceItem(None)

            if self.tmp_link_item.scene():
                self.scene.removeItem(self.tmp_link_item)

            self.tmp_link_item = None

        if self.current_target_item:
            if not self.showing_incompatible_widget:
                self.remove_tmp_anchor()
            else:
                if self.direction == self.FROM_SOURCE:
                    anchor = self.current_target_item.inputAnchorItem
                else:
                    anchor = self.current_target_item.outputAnchorItem
                anchor.setIncompatible(False)

            self.current_target_item = None

        if self.cursor_anchor_point and self.cursor_anchor_point.scene():
            self.scene.removeItem(self.cursor_anchor_point)
            self.cursor_anchor_point = None

    def reset_open_anchor(self):
        """
        This isn't part of cleanup, because it should retain its value
        until the link is created.
        """
        if self.direction == self.FROM_SOURCE:
            anchor = self.from_item.outputAnchorItem
        else:
            anchor = self.from_item.inputAnchorItem
        anchor.setKeepAnchorOpen(None)


def edit_links(
        scheme: Scheme,
        source_node: Node,
        sink_node: Node,
        initial_links: 'Optional[List[OIPair]]' = None,
        parent: 'Optional[QWidget]' = None
) -> 'Tuple[int, List[OIPair], List[OIPair]]':
    """
    Show and execute the `EditLinksDialog`.
    Optional `initial_links` list can provide a list of initial
    `(source, sink)` channel tuples to show in the view, otherwise
    the dialog is populated with existing links in the scheme (passing
    an empty list will disable all initial links).

    """
    log.info("Constructing a Link Editor dialog.")

    dlg = EditLinksDialog(parent, windowTitle="Edit Links")

    # all SchemeLinks between the two nodes.
    links = scheme.find_links(source_node=source_node, sink_node=sink_node)
    existing_links = [(link.source_channel, link.sink_channel)
                      for link in links]

    if initial_links is None:
        initial_links = list(existing_links)

    dlg.setNodes(source_node, sink_node)
    dlg.setLinks(initial_links)

    log.info("Executing a Link Editor Dialog.")
    rval = dlg.exec()

    if rval == EditLinksDialog.Accepted:
        edited_links = dlg.links()

        # Differences
        links_to_add = set(edited_links) - set(existing_links)
        links_to_remove = set(existing_links) - set(edited_links)
        return rval, list(links_to_add), list(links_to_remove)
    else:
        return rval, [], []


def add_links_plan(scheme, links, force_replace=False):
    # type: (Scheme, Iterable[Link], bool) -> Tuple[List[Link], List[Link]]
    """
    Return a plan for adding a list of links to the scheme.
    """
    links_to_add = list(links)
    links_to_remove = [conflicting_single_link(scheme, link)
                       for link in links]
    links_to_remove = [link for link in links_to_remove if link is not None]

    if not force_replace:
        links_to_add, links_to_remove = remove_duplicates(links_to_add,
                                                          links_to_remove)
    return links_to_add, links_to_remove


def conflicting_single_link(scheme, link):
    # type: (Scheme, Link) -> Optional[Link]
    """
    Find and return an existing link in `scheme` connected to the same
    input channel as `link` if the channel has the 'single' flag.
    If no such channel exists (or sink channel is not 'single')
    return `None`.
    """
    if link.sink_channel.single:
        existing = scheme.find_links(
            sink_node=link.sink_node,
            sink_channel=link.sink_channel
        )

        if existing:
            assert len(existing) == 1
            return existing[0]
    return None


def remove_duplicates(links_to_add, links_to_remove):
    # type: (List[Link], List[Link]) -> Tuple[List[Link], List[Link]]
    def link_key(link):
        # type: (Link) -> Tuple[Node, OutputSignal, Node, InputSignal]
        return (link.source_node, link.source_channel,
                link.sink_node, link.sink_channel)

    add_keys = list(map(link_key, links_to_add))
    remove_keys = list(map(link_key, links_to_remove))
    duplicate_keys = set(add_keys).intersection(remove_keys)

    def not_duplicate(link):
        # type: (Link) -> bool
        return link_key(link) not in duplicate_keys

    links_to_add = list(filter(not_duplicate, links_to_add))
    links_to_remove = list(filter(not_duplicate, links_to_remove))
    return links_to_add, links_to_remove


class NewNodeAction(UserInteraction):
    """
    Present the user with a quick menu for node selection and
    create the selected node.
    """
    def mousePressEvent(self, event):
        # type: (QGraphicsSceneMouseEvent) -> bool
        if event.button() == Qt.RightButton:
            self.create_new(event.screenPos())
            self.end()
        return True

    def create_new(self, pos, search_text=""):
        # type: (QPoint, str) -> Optional[Node]
        """
        Create and add new node to the workflow using `QuickMenu` popup at
        `pos` (in screen coordinates).
        """
        menu = self.document.quickMenu()
        menu.setFilterFunc(None)

        # compares probability of the user needing the widget as a source
        def defaultSort(left, right):
            default_suggestions = self.suggestions.get_default_suggestions()
            left_frequency = sum(default_suggestions[left].values())
            right_frequency = sum(default_suggestions[right].values())
            return left_frequency > right_frequency

        menu.setSortingFunc(defaultSort)

        action = menu.exec(pos, search_text)
        if action:
            item = action.property("item")
            desc = item.data(QtWidgetRegistry.WIDGET_DESC_ROLE)
            # Get the scene position
            view = self.document.view()
            pos = view.mapToScene(view.mapFromGlobal(pos))

            statistics = self.document.usageStatistics()
            statistics.begin_action(UsageStatistics.QuickMenu)
            node = self.document.newNodeHelper(desc,
                                               position=(pos.x(), pos.y()))
            self.document.addNode(node)
            return node
        else:
            return None


class RectangleSelectionAction(UserInteraction):
    """
    Select items in the scene using a Rectangle selection
    """
    def __init__(self, document, *args, **kwargs):
        # type: (SchemeEditWidget, Any, Any) -> None
        super().__init__(document, *args, **kwargs)
        # The initial selection at drag start
        self.initial_selection = None  # type: Optional[Set[QGraphicsItem]]
        # Selection when last updated in a mouseMoveEvent
        self.last_selection = None     # type: Optional[Set[QGraphicsItem]]
        # A selection rect (`QRectF`)
        self.selection_rect = None     # type: Optional[QRectF]
        # Keyboard modifiers
        self.modifiers = Qt.NoModifier
        self.rect_item = None          # type: Optional[QGraphicsRectItem]

    def mousePressEvent(self, event):
        # type: (QGraphicsSceneMouseEvent) -> bool
        pos = event.scenePos()
        any_item = self.scene.item_at(pos)
        if not any_item and event.button() & Qt.LeftButton:
            self.modifiers = event.modifiers()
            self.selection_rect = QRectF(pos, QSizeF(0, 0))
            self.rect_item = QGraphicsRectItem(
                self.selection_rect.normalized()
            )

            self.rect_item.setPen(
                QPen(QBrush(QColor(51, 153, 255, 192)),
                     0.4, Qt.SolidLine, Qt.RoundCap)
            )

            self.rect_item.setBrush(
                QBrush(QColor(168, 202, 236, 192))
            )

            self.rect_item.setZValue(-100)

            # Clear the focus if necessary.
            if not self.scene.stickyFocus():
                self.scene.clearFocus()

            if not self.modifiers & Qt.ControlModifier:
                self.scene.clearSelection()

            event.accept()
            return True
        else:
            self.cancel(self.ErrorReason)
            return False

    def mouseMoveEvent(self, event):
        # type: (QGraphicsSceneMouseEvent) -> bool
        if self.rect_item is not None and not self.rect_item.scene():
            # Add the rect item to the scene when the mouse moves.
            self.scene.addItem(self.rect_item)
        self.update_selection(event)
        return True

    def mouseReleaseEvent(self, event):
        # type: (QGraphicsSceneMouseEvent) -> bool
        if event.button() == Qt.LeftButton:
            if self.initial_selection is None:
                # A single click.
                self.scene.clearSelection()
            else:
                self.update_selection(event)
        self.end()
        return True

    def update_selection(self, event):
        # type: (QGraphicsSceneMouseEvent) -> None
        """
        Update the selection rectangle from a QGraphicsSceneMouseEvent
        `event` instance.
        """
        if self.initial_selection is None:
            self.initial_selection = set(self.scene.selectedItems())
            self.last_selection = self.initial_selection

        assert self.selection_rect is not None
        assert self.rect_item is not None
        assert self.initial_selection is not None
        assert self.last_selection is not None

        pos = event.scenePos()
        self.selection_rect = QRectF(self.selection_rect.topLeft(), pos)

        # Make sure the rect_item does not cause the scene rect to grow.
        rect = self._bound_selection_rect(self.selection_rect.normalized())

        # Need that 0.5 constant otherwise the sceneRect will still
        # grow (anti-aliasing correction by QGraphicsScene?)
        pw = self.rect_item.pen().width() + 0.5

        self.rect_item.setRect(rect.adjusted(pw, pw, -pw, -pw))

        selected = self.scene.items(self.selection_rect.normalized(),
                                    Qt.IntersectsItemShape,
                                    Qt.AscendingOrder)

        selected = set([item for item in selected if
                        item.flags() & QGraphicsItem.ItemIsSelectable])

        if self.modifiers & Qt.ControlModifier:
            for item in selected | self.last_selection | \
                    self.initial_selection:
                item.setSelected(
                    (item in selected) ^ (item in self.initial_selection)
                )
        else:
            for item in selected.union(self.last_selection):
                item.setSelected(item in selected)

        self.last_selection = set(self.scene.selectedItems())

    def end(self):
        # type: () -> None
        self.initial_selection = None
        self.last_selection = None
        self.modifiers = Qt.NoModifier
        if self.rect_item is not None:
            self.rect_item.hide()
            if self.rect_item.scene() is not None:
                self.scene.removeItem(self.rect_item)
        super().end()

    def viewport_rect(self):
        # type: () -> QRectF
        """
        Return the bounding rect of the document's viewport on the scene.
        """
        view = self.document.view()
        vsize = view.viewport().size()
        viewportrect = QRect(0, 0, vsize.width(), vsize.height())
        return view.mapToScene(viewportrect).boundingRect()

    def _bound_selection_rect(self, rect):
        # type: (QRectF) -> QRectF
        """
        Bound the selection `rect` to a sensible size.
        """
        srect = self.scene.sceneRect()
        vrect = self.viewport_rect()
        maxrect = srect.united(vrect)
        return rect.intersected(maxrect)


class EditNodeLinksAction(UserInteraction):
    """
    Edit multiple links between two :class:`SchemeNode` instances using
    a :class:`EditLinksDialog`

    Parameters
    ----------
    document : :class:`SchemeEditWidget`
        The editor widget.
    source_node : :class:`SchemeNode`
        The source (link start) node for the link editor.
    sink_node : :class:`SchemeNode`
        The sink (link end) node for the link editor.

    """
    def __init__(self, document, source_node, sink_node, *args, **kwargs):
        # type: (SchemeEditWidget, Node, Node, Any, Any) -> None
        super().__init__(document, *args, **kwargs)
        self.source_node = source_node
        self.sink_node = sink_node

    def edit_links(self, initial_links=None):
        # type: (Optional[List[OIPair]]) -> None
        """
        Show and execute the `EditLinksDialog`.
        Optional `initial_links` list can provide a list of initial
        `(source, sink)` channel tuples to show in the view, otherwise
        the dialog is populated with existing links in the scheme (passing
        an empty list will disable all initial links).

        """
        log.info("Constructing a Link Editor dialog.")

        dlg = EditLinksDialog(self.document, windowTitle="Edit Links")

        links = self.scheme.find_links(source_node=self.source_node,
                                       sink_node=self.sink_node)
        existing_links = [(link.source_channel, link.sink_channel)
                          for link in links]

        if initial_links is None:
            initial_links = list(existing_links)

        dlg.setNodes(self.source_node, self.sink_node)
        dlg.setLinks(initial_links)

        log.info("Executing a Link Editor Dialog.")
        rval = dlg.exec()

        if rval == EditLinksDialog.Accepted:
            links_spec = dlg.links()

            links_to_add = set(links_spec) - set(existing_links)
            links_to_remove = set(existing_links) - set(links_spec)

            stack = self.document.undoStack()
            stack.beginMacro("Edit Links")

            # First remove links into a 'Single' sink channel,
            # but only the ones that do not have self.source_node as
            # a source (they will be removed later from links_to_remove)
            for _, sink_channel in links_to_add:
                if sink_channel.single:
                    existing = self.scheme.find_links(
                        sink_node=self.sink_node,
                        sink_channel=sink_channel
                    )

                    existing = [link for link in existing
                                if link.source_node is not self.source_node]

                    if existing:
                        assert len(existing) == 1
                        self.document.removeLink(existing[0])

            for source_channel, sink_channel in links_to_remove:
                links = self.scheme.find_links(source_node=self.source_node,
                                               source_channel=source_channel,
                                               sink_node=self.sink_node,
                                               sink_channel=sink_channel)
                assert len(links) == 1
                self.document.removeLink(links[0])

            for source_channel, sink_channel in links_to_add:
                link = scheme.SchemeLink(self.source_node, source_channel,
                                         self.sink_node, sink_channel)

                self.document.addLink(link)

            stack.endMacro()


def point_to_tuple(point):
    # type: (QPointF) -> Tuple[float, float]
    """
    Convert a QPointF into a (x, y) tuple.
    """
    return (point.x(), point.y())


class NewArrowAnnotation(UserInteraction):
    """
    Create a new arrow annotation handler.
    """
    def __init__(self, document, *args, **kwargs):
        # type: (SchemeEditWidget, Any, Any) -> None
        super().__init__(document, *args, **kwargs)
        self.down_pos = None  # type: Optional[QPointF]
        self.arrow_item = None  # type: Optional[items.ArrowAnnotation]
        self.annotation = None  # type: Optional[scheme.SchemeArrowAnnotation]
        self.color = "red"
        self.cancelOnEsc = True

    def start(self):
        # type: () -> None
        self.document.view().setCursor(Qt.CrossCursor)

        helpevent = QuickHelpTipEvent(
            self.tr("Click and drag to create a new arrow"),
            self.tr('<h3>New arrow annotation</h3>'
                    '<p>Click and drag to create a new arrow annotation</p>'
#                    '<a href="help://orange-canvas/arrow-annotations>'
#                    'More ...</a>'
                    )
        )
        QCoreApplication.postEvent(self.document, helpevent)

        super().start()

    def setColor(self, color):
        """
        Set the color for the new arrow.
        """
        self.color = color

    def mousePressEvent(self, event):
        # type: (QGraphicsSceneMouseEvent) -> bool
        if event.button() == Qt.LeftButton:
            self.down_pos = event.scenePos()
            event.accept()
            return True
        else:
            return super().mousePressEvent(event)

    def mouseMoveEvent(self, event):
        # type: (QGraphicsSceneMouseEvent) -> bool
        if event.buttons() & Qt.LeftButton:
            assert self.down_pos is not None
            if self.arrow_item is None and \
                    (self.down_pos - event.scenePos()).manhattanLength() > \
                    QApplication.instance().startDragDistance():

                annot = scheme.SchemeArrowAnnotation(
                    point_to_tuple(self.down_pos),
                    point_to_tuple(event.scenePos())
                )
                annot.set_color(self.color)
                item = self.scene.add_annotation(annot)

                self.arrow_item = item
                self.annotation = annot

            if self.arrow_item is not None:
                p1, p2 = map(self.arrow_item.mapFromScene,
                             (self.down_pos, event.scenePos()))
                self.arrow_item.setLine(QLineF(p1, p2))

            event.accept()
            return True
        else:
            return super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event):
        # type: (QGraphicsSceneMouseEvent) -> bool
        if event.button() == Qt.LeftButton:
            if self.arrow_item is not None:
                assert self.down_pos is not None and self.annotation is not None
                p1, p2 = self.down_pos, event.scenePos()
                # Commit the annotation to the scheme
                self.annotation.set_line(point_to_tuple(p1),
                                         point_to_tuple(p2))

                self.document.addAnnotation(self.annotation)

                p1, p2 = map(self.arrow_item.mapFromScene, (p1, p2))
                self.arrow_item.setLine(QLineF(p1, p2))

            self.end()
            return True
        else:
            return super().mouseReleaseEvent(event)

    def cancel(self, reason=UserInteraction.OtherReason):  # type: (int) -> None
        if self.arrow_item is not None:
            self.scene.removeItem(self.arrow_item)
            self.arrow_item = None
        super().cancel(reason)

    def end(self):
        # type: () -> None
        self.down_pos = None
        self.arrow_item = None
        self.annotation = None
        self.document.view().setCursor(Qt.ArrowCursor)

        # Clear the help tip
        helpevent = QuickHelpTipEvent("", "")
        QCoreApplication.postEvent(self.document, helpevent)

        super().end()


def rect_to_tuple(rect):
    # type: (QRectF) -> Tuple[float, float, float, float]
    """
    Convert a QRectF into a (x, y, width, height) tuple.
    """
    return rect.x(), rect.y(), rect.width(), rect.height()


class NewTextAnnotation(UserInteraction):
    """
    A New Text Annotation interaction handler
    """
    def __init__(self, document, *args, **kwargs):
        # type: (SchemeEditWidget, Any, Any) -> None
        super().__init__(document, *args, **kwargs)
        self.down_pos = None  # type: Optional[QPointF]
        self.annotation_item = None  # type: Optional[items.TextAnnotation]
        self.annotation = None  # type: Optional[scheme.SchemeTextAnnotation]
        self.control = None  # type: Optional[controlpoints.ControlPointRect]
        self.font = document.font()  # type: QFont
        self.cancelOnEsc = True

    def setFont(self, font):
        # type: (QFont) -> None
        self.font = QFont(font)

    def start(self):
        # type: () -> None
        self.document.view().setCursor(Qt.CrossCursor)

        helpevent = QuickHelpTipEvent(
            self.tr("Click to create a new text annotation"),
            self.tr('<h3>New text annotation</h3>'
                    '<p>Click (and drag to resize) on the canvas to create '
                    'a new text annotation item.</p>'
#                    '<a href="help://orange-canvas/text-annotations">'
#                    'More ...</a>'
                    )
        )
        QCoreApplication.postEvent(self.document, helpevent)

        super().start()

    def createNewAnnotation(self, rect):
        # type: (QRectF) -> None
        """
        Create a new TextAnnotation at with `rect` as the geometry.
        """
        annot = scheme.SchemeTextAnnotation(rect_to_tuple(rect))
        font = {"family": self.font.family(),
                "size": self.font.pixelSize()}
        annot.set_font(font)

        item = self.scene.add_annotation(annot)
        item.setTextInteractionFlags(Qt.TextEditorInteraction)
        item.setFramePen(QPen(Qt.DashLine))

        self.annotation_item = item
        self.annotation = annot
        self.control = controlpoints.ControlPointRect()
        self.control.rectChanged.connect(item.setGeometry)
        self.scene.addItem(self.control)

    def mousePressEvent(self, event):
        # type: (QGraphicsSceneMouseEvent) -> bool
        if event.button() == Qt.LeftButton:
            self.down_pos = event.scenePos()
            return True
        return super().mousePressEvent(event)

    def mouseMoveEvent(self, event):
        # type: (QGraphicsSceneMouseEvent) -> bool
        if event.buttons() & Qt.LeftButton:
            assert self.down_pos is not None
            if self.annotation_item is None and \
                    (self.down_pos - event.scenePos()).manhattanLength() > \
                    QApplication.instance().startDragDistance():
                rect = QRectF(self.down_pos, event.scenePos()).normalized()
                self.createNewAnnotation(rect)

            if self.annotation_item is not None:
                assert self.control is not None
                rect = QRectF(self.down_pos, event.scenePos()).normalized()
                self.control.setRect(rect)
            return True
        return super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event):
        # type: (QGraphicsSceneMouseEvent) -> bool
        if event.button() == Qt.LeftButton:
            if self.annotation_item is None:
                self.createNewAnnotation(QRectF(event.scenePos(),
                                                event.scenePos()))
                rect = self.defaultTextGeometry(event.scenePos())
            else:
                assert self.down_pos is not None
                rect = QRectF(self.down_pos, event.scenePos()).normalized()
            assert self.annotation_item is not None
            assert self.control is not None
            assert self.annotation is not None
            # Commit the annotation to the scheme.
            self.annotation.rect = rect_to_tuple(rect)

            self.document.addAnnotation(self.annotation)

            self.annotation_item.setGeometry(rect)

            self.control.rectChanged.disconnect(
                self.annotation_item.setGeometry
            )
            self.control.hide()

            # Move the focus to the editor.
            self.annotation_item.setFramePen(QPen(Qt.NoPen))
            self.annotation_item.setFocus(Qt.OtherFocusReason)
            self.annotation_item.startEdit()

            self.end()
            return True
        return super().mouseMoveEvent(event)

    def defaultTextGeometry(self, point):
        # type: (QPointF) -> QRectF
        """
        Return the default text geometry. Used in case the user single
        clicked in the scene.
        """
        assert self.annotation_item is not None
        font = self.annotation_item.font()
        metrics = QFontMetrics(font)
        spacing = metrics.lineSpacing()
        margin = self.annotation_item.document().documentMargin()

        rect = QRectF(QPointF(point.x(), point.y() - spacing - margin),
                      QSizeF(150, spacing + 2 * margin))
        return rect

    def cancel(self, reason=UserInteraction.OtherReason):  # type: (int) -> None
        if self.annotation_item is not None:
            self.annotation_item.clearFocus()
            self.scene.removeItem(self.annotation_item)
            self.annotation_item = None
        super().cancel(reason)

    def end(self):
        # type: () -> None
        if self.control is not None:
            self.scene.removeItem(self.control)

        self.control = None
        self.down_pos = None
        self.annotation_item = None
        self.annotation = None
        self.document.view().setCursor(Qt.ArrowCursor)

        # Clear the help tip
        helpevent = QuickHelpTipEvent("", "")
        QCoreApplication.postEvent(self.document, helpevent)

        super().end()


class ResizeTextAnnotation(UserInteraction):
    """
    Resize a Text Annotation interaction handler.
    """
    def __init__(self, document, *args, **kwargs):
        # type: (SchemeEditWidget, Any, Any) -> None
        super().__init__(document, *args, **kwargs)
        self.item = None        # type: Optional[items.TextAnnotation]
        self.annotation = None  # type: Optional[scheme.SchemeTextAnnotation]
        self.control = None     # type: Optional[controlpoints.ControlPointRect]
        self.savedFramePen = None  # type: Optional[QPen]
        self.savedRect = None      # type: Optional[QRectF]

    def mousePressEvent(self, event):
        # type: (QGraphicsSceneMouseEvent) -> bool
        pos = event.scenePos()
        if event.button() & Qt.LeftButton and self.item is None:
            item = self.scene.item_at(pos, items.TextAnnotation)
            if item is not None and not item.hasFocus():
                self.editItem(item)
                return False
        return super().mousePressEvent(event)

    def editItem(self, item):
        # type: (items.TextAnnotation) -> None
        annotation = self.scene.annotation_for_item(item)
        rect = item.geometry()  # TODO: map to scene if item has a parent.
        control = controlpoints.ControlPointRect(rect=rect)
        self.scene.addItem(control)

        self.savedFramePen = item.framePen()
        self.savedRect = rect

        control.rectEdited.connect(item.setGeometry)
        control.setFocusProxy(item)

        item.setFramePen(QPen(Qt.DashDotLine))
        item.geometryChanged.connect(self.__on_textGeometryChanged)

        self.item = item

        self.annotation = annotation
        self.control = control

    def commit(self):
        # type: () -> None
        """
        Commit the current item geometry state to the document.
        """
        if self.item is None:
            return
        rect = self.item.geometry()
        if self.savedRect != rect:
            command = commands.SetAttrCommand(
                self.annotation, "rect",
                (rect.x(), rect.y(), rect.width(), rect.height()),
                name="Edit text geometry"
            )
            self.document.undoStack().push(command)
            self.savedRect = rect

    def __on_editingFinished(self):
        # type: () -> None
        self.commit()
        self.end()

    def __on_rectEdited(self, rect):
        # type: (QRectF) -> None
        assert self.item is not None
        self.item.setGeometry(rect)

    def __on_textGeometryChanged(self):
        # type: () -> None
        assert self.control is not None and self.item is not None
        if not self.control.isControlActive():
            rect = self.item.geometry()
            self.control.setRect(rect)

    def cancel(self, reason=UserInteraction.OtherReason):
        # type: (int) -> None
        log.debug("ResizeTextAnnotation.cancel(%s)", reason)
        if self.item is not None and self.savedRect is not None:
            self.item.setGeometry(self.savedRect)
        super().cancel(reason)

    def end(self):
        # type: () -> None
        if self.control is not None:
            self.scene.removeItem(self.control)

        if self.item is not None and self.savedFramePen is not None:
            self.item.setFramePen(self.savedFramePen)

        self.item = None
        self.annotation = None
        self.control = None

        super().end()


class ResizeArrowAnnotation(UserInteraction):
    """
    Resize an Arrow Annotation interaction handler.
    """
    def __init__(self, document, *args, **kwargs):
        # type: (SchemeEditWidget, Any, Any) -> None
        super().__init__(document, *args, **kwargs)
        self.item = None        # type: Optional[items.ArrowAnnotation]
        self.annotation = None  # type: Optional[scheme.SchemeArrowAnnotation]
        self.control = None     # type: Optional[controlpoints.ControlPointLine]
        self.savedLine = None   # type: Optional[QLineF]

    def mousePressEvent(self, event):
        # type: (QGraphicsSceneMouseEvent) -> bool
        pos = event.scenePos()
        if self.item is None:
            item = self.scene.item_at(pos, items.ArrowAnnotation)
            if item is not None and not item.hasFocus():
                self.editItem(item)
                return False

        return super().mousePressEvent(event)

    def editItem(self, item):
        # type: (items.ArrowAnnotation) -> None
        annotation = self.scene.annotation_for_item(item)
        control = controlpoints.ControlPointLine()
        self.scene.addItem(control)

        line = item.line()
        self.savedLine = line

        p1, p2 = map(item.mapToScene, (line.p1(), line.p2()))

        control.setLine(QLineF(p1, p2))
        control.setFocusProxy(item)
        control.lineEdited.connect(self.__on_lineEdited)

        item.geometryChanged.connect(self.__on_lineGeometryChanged)

        self.item = item
        self.annotation = annotation
        self.control = control

    def commit(self):
        # type: () -> None
        """Commit the current geometry of the item to the document.

        Does nothing if the actual geometry has not changed.
        """
        if self.control is None or self.item is None:
            return
        line = self.control.line()
        p1, p2 = line.p1(), line.p2()

        if self.item.line() != self.savedLine:
            command = commands.SetAttrCommand(
                self.annotation,
                "geometry",
                ((p1.x(), p1.y()), (p2.x(), p2.y())),
                name="Edit arrow geometry",
            )
            self.document.undoStack().push(command)
            self.savedLine = self.item.line()

    def __on_editingFinished(self):
        # type: () -> None
        self.commit()
        self.end()

    def __on_lineEdited(self, line):
        # type: (QLineF) -> None
        if self.item is not None:
            p1, p2 = map(self.item.mapFromScene, (line.p1(), line.p2()))
            self.item.setLine(QLineF(p1, p2))

    def __on_lineGeometryChanged(self):
        # type: () -> None
        # Possible geometry change from out of our control, for instance
        # item move as a part of a selection group.
        assert self.control is not None and self.item is not None
        if not self.control.isControlActive():
            assert self.item is not None
            line = self.item.line()
            p1, p2 = map(self.item.mapToScene, (line.p1(), line.p2()))
            self.control.setLine(QLineF(p1, p2))

    def cancel(self, reason=UserInteraction.OtherReason):
        # type: (int) -> None
        log.debug("ResizeArrowAnnotation.cancel(%s)", reason)
        if self.item is not None and self.savedLine is not None:
            self.item.setLine(self.savedLine)

        super().cancel(reason)

    def end(self):
        # type: () -> None
        if self.control is not None:
            self.scene.removeItem(self.control)

        if self.item is not None:
            self.item.geometryChanged.disconnect(self.__on_lineGeometryChanged)

        self.control = None
        self.item = None
        self.annotation = None

        super().end()


class DropHandler(abc.ABC):
    """
    An abstract drop handler.

    .. versionadded:: 0.1.20
    """
    @abc.abstractmethod
    def accepts(self, document: 'SchemeEditWidget', event: 'QGraphicsSceneDragDropEvent') -> bool:
        """
        Returns True if a `document` can accept a drop of the data from `event`.
        """
        return False

    @abc.abstractmethod
    def doDrop(self, document: 'SchemeEditWidget', event: 'QGraphicsSceneDragDropEvent') -> bool:
        """
        Complete the drop of data from `event` onto the `document`.
        """
        return False


class DropHandlerAction(abc.ABC):
    @abc.abstractmethod
    def actionFromDropEvent(
            self, document: 'SchemeEditWidget', event: 'QGraphicsSceneDragDropEvent'
    ) -> QAction:
        """
        Create and return an QAction representing a drop action.

        This action is used to disambiguate between possible drop actions.

        The action can have sub menus, however all actions in submenus **must**
        have the `DropHandler` instance set as their `QAction.data()`.

        The actions **must not** execute the actual drop from their triggered
        slot connections. The drop will be dispatched to the `action.data()`
        handler's `doDrop()` after that action is triggered and the menu is
        closed.
        """
        raise NotImplementedError


class NodeFromMimeDataDropHandler(DropHandler, DropHandlerAction):
    """
    Create a new node from dropped mime data.

    Subclasses must override `canDropMimeData`, `parametersFromMimeData`,
    and `qualifiedName`.

    .. versionadded:: 0.1.20
    """
    @abc.abstractmethod
    def qualifiedName(self) -> str:
        """
        The qualified name for the node created by this handler. The handler
        will not be invoked if this name does not appear in the registry
        associated with the workflow.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def canDropMimeData(self, document: 'SchemeEditWidget', data: 'QMimeData') -> bool:
        """
        Can the handler create a node from the drop mime data.

        Reimplement this in a subclass to check if the `data` has appropriate
        format.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def parametersFromMimeData(self, document: 'SchemeEditWidget', data: 'QMimeData') -> 'Dict[str, Any]':
        """
        Return the node parameters based from the drop mime data.
        """
        raise NotImplementedError

    def accepts(self, document: 'SchemeEditWidget', event: 'QGraphicsSceneDragDropEvent') -> bool:
        """Reimplemented."""
        reg = document.registry()
        if not reg.has_widget(self.qualifiedName()):
            return False
        return self.canDropMimeData(document, event.mimeData())

    def nodeFromMimeData(self, document: 'SchemeEditWidget', data: 'QMimeData') -> 'Node':
        reg = document.registry()
        wd = reg.widget(self.qualifiedName())
        node = document.newNodeHelper(wd)
        parameters = self.parametersFromMimeData(document, data)
        node.properties = parameters
        return node

    def doDrop(self, document: 'SchemeEditWidget', event: 'QGraphicsSceneDragDropEvent') -> bool:
        """Reimplemented."""
        reg = document.registry()
        if not reg.has_widget(self.qualifiedName()):
            return False
        node = self.nodeFromMimeData(document, event.mimeData())
        node.position = (event.scenePos().x(), event.scenePos().y())
        activate = self.shouldActivateNode()
        wd = document.widgetManager()
        if activate and wd is not None:
            def activate(node_, widget):
                if node_ is node:
                    try:
                        self.activateNode(document, node, widget)
                    finally:
                        # self-disconnect the slot
                        wd.widget_for_node_added.disconnect(activate)
            wd.widget_for_node_added.connect(activate)
        document.addNode(node)
        if activate:
            QApplication.postEvent(node, WorkflowEvent(WorkflowEvent.NodeActivateRequest))
        return True

    def shouldActivateNode(self) -> bool:
        """
        Should the new dropped node activate (open GUI controller) immediately.

        If this method returns `True` then the `activateNode` method will be
        called after the node has been added and the GUI controller created.

        The default implementation returns False.
        """
        return False

    def activateNode(self, document: 'SchemeEditWidget', node: 'Node', widget: 'QWidget') -> None:
        """
        Activate (open) the `node`'s GUI controller `widget` after a
        completed drop.

        Reimplement this if the node requires further configuration via the
        GUI.

        The default implementation delegates to the :class:`WidgetManager`
        associated with the document.
        """
        wd = document.widgetManager()
        if wd is not None:
            wd.activate_widget_for_node(node, widget)
        else:
            widget.show()

    def actionFromDropEvent(
            self, document: 'SchemeEditWidget', event: 'QGraphicsSceneDragDropEvent'
    ) -> QAction:
        """Reimplemented."""
        reg = document.registry()
        ac = QAction(None)
        ac.setData(self)
        if reg is not None:
            desc = reg.widget(self.qualifiedName())
            ac.setText(desc.name)
            ac.setToolTip(tooltip_helper(desc))
            ac.setWhatsThis(whats_this_helper(desc))
        else:
            ac.setText(f"{self.qualifiedName()}")
            ac.setEnabled(False)
            ac.setVisible(False)
        return ac


def load_entry_point(
        ep: EntryPoint, log: logging.Logger = None,
) -> Tuple['EntryPoint', Any]:
    if log is None:
        log = logging.getLogger(__name__)
    try:
        value = ep.load()
    except (ImportError, AttributeError):
        log.exception("Could not load %s", ep)
    except Exception:  # noqa
        log.exception("Unexpected Error; %s will be skipped", ep)
    else:
        return ep, value


def iter_load_entry_points(
        iter: Iterable[EntryPoint], log: logging.Logger = None,
):
    if log is None:
        log = logging.getLogger(__name__)
    for ep in iter:
        try:
            ep, value = load_entry_point(ep, log)
        except Exception:
            pass
        else:
            yield ep, value


class PluginDropHandler(DropHandler):
    """
    Delegate drop event processing to plugin drop handlers.

    .. versionadded:: 0.1.20
    """
    #: The default entry point group
    ENTRY_POINT = "orangecanvas.document.interactions.DropHandler"

    def __init__(self, group=ENTRY_POINT, **kwargs):
        super().__init__(**kwargs)
        self.__group = group

    def iterEntryPoints(self) -> Iterable['EntryPoint']:
        """
        Return an iterator over all entry points.
        """
        eps = entry_points().get(self.__group, [])
        # Can have duplicated entries here if a distribution is *found* via
        # different `sys.meta_path` handlers and/or on different `sys.path`
        # entries.
        return unique(eps, key=lambda ep: ep.value)

    __entryPoints = None

    def entryPoints(self) -> Iterable[Tuple['EntryPoint', 'DropHandler']]:
        """
        Return an iterator over entry points and instantiated drop handlers.
        """
        eps = []
        if self.__entryPoints:
            ep_iter = self.__entryPoints
            store_eps = lambda ep, value: None
        else:
            ep_iter = self.iterEntryPoints()
            ep_iter = iter_load_entry_points(ep_iter, log)
            store_eps = lambda ep, value: eps.append((ep, value))

        for ep, value in ep_iter:
            if not issubclass(value, DropHandler):
                log.error(
                    f"{ep} yielded {type(value)}, expected a "
                    f"{DropHandler} subtype"
                )
                continue
            try:
                handler = value()
            except Exception:  # noqa
                log.exception("Error in default constructor of %s", value)
            else:
                yield ep, handler
                store_eps(ep, value)

        self.__entryPoints = tuple(eps)

    __accepts: Sequence[Tuple[EntryPoint, DropHandler]] = ()

    def accepts(self, document: 'SchemeEditWidget', event: 'QGraphicsSceneDragDropEvent') -> bool:
        """
        Reimplemented.

        Accept the event if any plugin handlers accept the event.
        """
        accepts = []
        self.__accepts = ()
        for ep, handler in self.entryPoints():
            if handler.accepts(document, event):
                accepts.append((ep, handler))
        self.__accepts = tuple(accepts)
        return bool(accepts)

    def doDrop(
            self, document: 'SchemeEditWidget', event: 'QGraphicsSceneDragDropEvent'
    ) -> bool:
        """
        Reimplemented.

        Dispatch the drop to the plugin handler that accepted the event.
        In case there are multiple handlers that accepted the event, a menu
        is used to select the handler.
        """
        handler: Optional[DropHandler] = None
        if len(self.__accepts) == 1:
            ep, handler = self.__accepts[0]
        elif len(self.__accepts) > 1:
            menu = QMenu(event.widget())
            for ep_, handler_ in self.__accepts:
                ac = action_for_handler(handler_, document, event)
                if ac is None:
                    ac = menu.addAction(ep_.name, )
                else:
                    menu.addAction(ac)
                    ac.setParent(menu)
                if not ac.toolTip():
                    ac.setToolTip(f"{ep_.name} ({ep_.module_name})")
                ac.setData(handler_)
            action = menu.exec(event.screenPos())
            if action is not None:
                handler = action.data()
        if handler is None:
            return False
        return handler.doDrop(document, event)


def action_for_handler(handler: DropHandler, document, event) -> Optional[QAction]:
    if isinstance(handler, DropHandlerAction):
        return handler.actionFromDropEvent(document, event)
    else:
        return None


class DropAction(UserInteraction):
    """
    A drop action on the workflow.
    """
    def __init__(
            self, document, *args, dropHandlers: Sequence[DropHandler] = (),
            **kwargs
    ) -> None:
        super().__init__(document, *args, **kwargs)
        self.__designatedAction: Optional[DropHandler] = None
        self.__dropHandlers = dropHandlers

    def dropHandlers(self) -> Iterable[DropHandler]:
        """Return an iterable over drop handlers."""
        return iter(self.__dropHandlers)

    def canHandleDrop(self, event: 'QGraphicsSceneDragDropEvent') -> bool:
        """
        Can this interactions handle the drop `event`.

        The default implementation checks each `dropHandler` if it
        :func:`~DropHandler.accepts` the event. The first such handler that
        accepts is selected to be the designated handler and will receive
        the drop (:func:`~DropHandler.doDrop`).
        """
        for ep in self.dropHandlers():
            if ep.accepts(self.document, event):
                self.__designatedAction = ep
                return True
        else:
            return False

    def dragEnterEvent(self, event):
        if self.canHandleDrop(event):
            event.acceptProposedAction()
            return True
        else:
            return False

    def dragMoveEvent(self, event):
        if self.canHandleDrop(event):
            event.acceptProposedAction()
            return True
        else:
            return False

    def dragLeaveEvent(self, even):
        self.__designatedAction = None
        self.end()
        return False

    def dropEvent(self, event):
        if self.__designatedAction is not None:
            try:
                res = self.__designatedAction.doDrop(self.document, event)
            except Exception:  # noqa
                log.exception("")
                res = False
            if res:
                event.acceptProposedAction()
            self.end()
            return True
        else:
            self.end()
            return False



"""
Orange Canvas Tool Dock widget

"""
import sys
import warnings
from typing import Optional, Any

from AnyQt.QtWidgets import (
    QWidget, QSplitter, QVBoxLayout, QAction, QSizePolicy, QApplication,
    QToolButton, QTreeView
)
from AnyQt.QtGui import (
    QPalette, QBrush, QDrag, QResizeEvent, QHideEvent, QPaintEvent
)
from AnyQt.QtCore import (
    Qt, QSize, QObject, QPropertyAnimation, QEvent, QRect, QPoint,
    QAbstractItemModel, QModelIndex, QPersistentModelIndex, QEventLoop,
    QMimeData
)
from AnyQt.QtCore import pyqtProperty as Property, pyqtSignal as Signal

from .. import styles
from ..gui.iconengine import StyledIconEngine
from ..gui.toolgrid import ToolGrid, ToolGridButton
from ..gui.toolbar import DynamicResizeToolBar
from ..gui.quickhelp import QuickHelp
from ..gui.framelesswindow import FramelessWindow
from ..gui.utils import create_css_gradient, available_screen_geometry
from ..document.quickmenu import MenuPage
from .widgettoolbox import WidgetToolBox, iter_index, item_text, item_icon, item_tooltip
from ..registry.qt import QtWidgetRegistry


class SplitterResizer(QObject):
    """
    An object able to control the size of a widget in a QSplitter instance.
    """
    def __init__(self, parent=None, **kwargs):
        # type: (Optional[QObject], Any) -> None
        super().__init__(parent, **kwargs)
        self.__splitter = None  # type: Optional[QSplitter]
        self.__widget = None    # type: Optional[QWidget]
        self.__updateOnShow = True  # Need __update on next show event
        self.__animationEnabled = True
        self.__size = -1
        self.__expanded = False
        self.__animation = QPropertyAnimation(
            self, b"size_", self, duration=200
        )
        self.__action = QAction("toggle-expanded", self, checkable=True)
        self.__action.triggered[bool].connect(self.setExpanded)

    def setSize(self, size):
        # type: (int) -> None
        """
        Set the size of the controlled widget (either width or height
        depending on the orientation).

        .. note::
            The controlled widget's size is only updated when it it is shown.
        """
        if self.__size != size:
            self.__size = size
            self.__update()

    def size(self):
        # type: () -> int
        """
        Return the size of the widget in the splitter (either height of
        width) depending on the splitter orientation.
        """
        if self.__splitter and self.__widget:
            index = self.__splitter.indexOf(self.__widget)
            sizes = self.__splitter.sizes()
            return sizes[index]
        else:
            return -1

    size_ = Property(int, fget=size, fset=setSize)

    def setAnimationEnabled(self, enable):
        # type: (bool) -> None
        """Enable/disable animation."""
        self.__animation.setDuration(0 if enable else 200)

    def animationEnabled(self):
        # type: () -> bool
        return self.__animation.duration() == 0

    def setSplitterAndWidget(self, splitter, widget):
        # type: (QSplitter, QWidget) -> None
        """Set the QSplitter and QWidget instance the resizer should control.

        .. note:: the widget must be in the splitter.
        """
        if splitter and widget and not splitter.indexOf(widget) > 0:
            raise ValueError("Widget must be in a splitter.")

        if self.__widget is not None:
            self.__widget.removeEventFilter(self)
        if self.__splitter is not None:
            self.__splitter.removeEventFilter(self)

        self.__splitter = splitter
        self.__widget = widget

        if widget is not None:
            widget.installEventFilter(self)
        if splitter is not None:
            splitter.installEventFilter(self)

        self.__update()

        size = self.size()
        if self.__expanded and size == 0:
            self.open()
        elif not self.__expanded and size > 0:
            self.close()

    def toggleExpandedAction(self):
        # type: () -> QAction
        """Return a QAction that can be used to toggle expanded state.
        """
        return self.__action

    def toogleExpandedAction(self):
        warnings.warn(
            "'toogleExpandedAction is deprecated, use 'toggleExpandedAction' "
            "instead.", DeprecationWarning, stacklevel=2
        )
        return self.toggleExpandedAction()

    def open(self):
        # type: () -> None
        """Open the controlled widget (expand it to sizeHint).
        """
        self.__expanded = True
        self.__action.setChecked(True)

        if self.__splitter is None or self.__widget is None:
            return

        hint = self.__widget.sizeHint()

        if self.__splitter.orientation() == Qt.Vertical:
            end = hint.height()
        else:
            end = hint.width()

        self.__animation.setStartValue(0)
        self.__animation.setEndValue(end)
        self.__animation.start()

    def close(self):
        # type: () -> None
        """Close the controlled widget (shrink to size 0).
        """
        self.__expanded = False
        self.__action.setChecked(False)

        if self.__splitter is None or self.__widget is None:
            return

        self.__animation.setStartValue(self.size())
        self.__animation.setEndValue(0)
        self.__animation.start()

    def setExpanded(self, expanded):
        # type: (bool) -> None
        """Set the expanded state."""
        if self.__expanded != expanded:
            if expanded:
                self.open()
            else:
                self.close()

    def expanded(self):
        # type: () -> bool
        """Return the expanded state."""
        return self.__expanded

    def __update(self):
        # type: () -> None
        """Update the splitter sizes."""
        if self.__splitter and self.__widget:
            if sum(self.__splitter.sizes()) == 0:
                # schedule update on next show event
                self.__updateOnShow = True
                return

            splitter = self.__splitter
            index = splitter.indexOf(self.__widget)
            sizes = splitter.sizes()
            current = sizes[index]
            diff = current - self.__size
            sizes[index] = self.__size
            sizes[index - 1] = sizes[index - 1] + diff
            self.__splitter.setSizes(sizes)

    def eventFilter(self, obj, event):
        # type: (QObject, QEvent) -> bool
        if event.type() == QEvent.Resize and obj is self.__widget and \
                self.__animation.state() == QPropertyAnimation.Stopped:
            # Update the expanded state when the user opens/closes the widget
            # by dragging the splitter handle.
            assert self.__splitter is not None
            assert isinstance(event, QResizeEvent)
            if self.__splitter.orientation() == Qt.Vertical:
                size = event.size().height()
            else:
                size = event.size().width()

            if self.__expanded and size == 0:
                self.__action.setChecked(False)
                self.__expanded = False
            elif not self.__expanded and size > 0:
                self.__action.setChecked(True)
                self.__expanded = True

        if event.type() == QEvent.Show and obj is self.__splitter and \
                self.__updateOnShow:
            # Update the splitter state after receiving valid geometry
            self.__updateOnShow = False
            self.__update()
        return super().eventFilter(obj, event)


class QuickHelpWidget(QuickHelp):
    def minimumSizeHint(self):
        # type: () -> QSize
        """Reimplemented to allow the Splitter to resize the widget
        with a continuous animation.
        """
        hint = super().minimumSizeHint()
        return QSize(hint.width(), 0)


class CanvasToolDock(QWidget):
    """Canvas dock widget with widget toolbox, quick help and
    canvas actions.
    """
    def __init__(self, parent=None, **kwargs):
        # type: (Optional[QWidget], Any) -> None
        super().__init__(parent, **kwargs)

        self.__setupUi()

    def __setupUi(self):
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)

        self.toolbox = WidgetToolBox()

        self.help = QuickHelpWidget(objectName="quick-help")

        self.__splitter = QSplitter()
        self.__splitter.setOrientation(Qt.Vertical)

        self.__splitter.addWidget(self.toolbox)
        self.__splitter.addWidget(self.help)

        self.toolbar = DynamicResizeToolBar()
        self.toolbar.setMovable(False)
        self.toolbar.setFloatable(False)

        self.toolbar.setSizePolicy(QSizePolicy.Ignored,
                                   QSizePolicy.Preferred)

        layout.addWidget(self.__splitter, 10)
        layout.addWidget(self.toolbar)

        self.setLayout(layout)
        self.__splitterResizer = SplitterResizer(self)
        self.__splitterResizer.setSplitterAndWidget(self.__splitter, self.help)

    def setQuickHelpVisible(self, state):
        # type: (bool) -> None
        """Set the quick help box visibility status."""
        self.__splitterResizer.setExpanded(state)

    def quickHelpVisible(self):
        # type: () -> bool
        return self.__splitterResizer.expanded()

    def setQuickHelpAnimationEnabled(self, enabled):
        # type: (bool) -> None
        """Enable/disable the quick help animation."""
        self.__splitterResizer.setAnimationEnabled(enabled)

    def toggleQuickHelpAction(self):
        # type: () -> QAction
        """Return a checkable QAction for help show/hide."""
        return self.__splitterResizer.toggleExpandedAction()

    def toogleQuickHelpAction(self):
        warnings.warn(
            "'toogleQuickHelpAction' is deprecated, use "
            "'toggleQuickHelpAction' instead.", DeprecationWarning,
            stacklevel=2
        )
        return self.toggleQuickHelpAction()


class _ToolGridButton(ToolGridButton):
    def paintEvent(self, event: QPaintEvent) -> None:
        with StyledIconEngine.setOverridePalette(styles.breeze_light()):
            super().paintEvent(event)


class QuickCategoryToolbar(ToolGrid):
    """A toolbar with category buttons."""
    def __init__(self, parent=None, buttonSize=QSize(), iconSize=QSize(),
                 **kwargs):
        # type: (Optional[QWidget], QSize, QSize, Any) -> None
        super().__init__(parent, 1, buttonSize, iconSize,
                         Qt.ToolButtonIconOnly, **kwargs)
        self.__model = None  # type: Optional[QAbstractItemModel]

    def setColumnCount(self, count):
        raise Exception("Cannot set the column count on a Toolbar")

    def setModel(self, model):
        # type: (Optional[QAbstractItemModel]) -> None
        """
        Set the registry model.
        """
        if self.__model is not None:
            self.__model.dataChanged.disconnect(self.__on_dataChanged)
            self.__model.rowsInserted.disconnect(self.__on_rowsInserted)
            self.__model.rowsRemoved.disconnect(self.__on_rowsRemoved)
            self.clear()

        self.__model = model
        if model is not None:
            model.dataChanged.connect(self.__on_dataChanged)
            model.rowsInserted.connect(self.__on_rowsInserted)
            model.rowsRemoved.connect(self.__on_rowsRemoved)
            self.__initFromModel(model)

    def __initFromModel(self, model):
        # type: (QAbstractItemModel) -> None
        """
        Initialize the toolbar from the model.
        """
        for index in iter_index(model, QModelIndex()):
            action = self.createActionForItem(index)
            self.addAction(action)

    def createActionForItem(self, index):
        # type: (QModelIndex) -> QAction
        """
        Create the QAction instance for item at `index` (`QModelIndex`).
        """
        action = QAction(
            item_icon(index), item_text(index), self,
            toolTip=item_tooltip(index)
        )
        action.setData(QPersistentModelIndex(index))
        return action

    def createButtonForAction(self, action):
        # type: (QAction) -> QToolButton
        """
        Create a button for the action.
        """
        button = _ToolGridButton(
            self,
            toolButtonStyle=self.toolButtonStyle(),
            iconSize=self.iconSize(),
        )
        button.setDefaultAction(action)
        if self.buttonSize().isValid():
            button.setFixedSize(self.buttonSize())
        item = action.data()  # QPersistentModelIndex
        assert isinstance(item, QPersistentModelIndex)

        brush = item.data(Qt.BackgroundRole)
        if not isinstance(brush, QBrush):
            brush = item.data(QtWidgetRegistry.BACKGROUND_ROLE)
            if not isinstance(brush, QBrush):
                brush = self.palette().brush(QPalette.Button)

        palette = button.palette()
        palette.setColor(QPalette.Button, brush.color())
        palette.setColor(QPalette.Window, brush.color())
        button.setPalette(palette)
        button.setProperty("quick-category-toolbutton", True)

        style_sheet = ("QToolButton {\n"
                       "    background: %s;\n"
                       "    border: none;\n"
                       "    border-bottom: 1px solid palette(mid);\n"
                       "}")
        button.setStyleSheet(style_sheet % create_css_gradient(brush.color()))

        return button

    def __on_dataChanged(self, topLeft, bottomRight):
        # type: (QModelIndex, QModelIndex) -> None
        assert self.__model is not None
        parent = topLeft.parent()
        if not parent.isValid():
            for row in range(topLeft.row(), bottomRight.row() + 1):
                item = self.__model.index(row, 0)
                action = self.actions()[row]
                action.setText(item_text(item))
                action.setIcon(item_icon(item))
                action.setToolTip(item_tooltip(item))

    def __on_rowsInserted(self, parent, start, end):
        # type: (QModelIndex, int, int) -> None
        assert self.__model is not None
        if not parent.isValid():
            for row in range(start, end + 1):
                item = self.__model.index(row, 0)
                self.insertAction(row, self.createActionForItem(item))

    def __on_rowsRemoved(self, parent, start, end):
        # type: (QModelIndex, int, int) -> None
        assert self.__model is not None
        if not parent.isValid():
            for row in range(end, start - 1, -1):
                action = self.actions()[row]
                self.removeAction(action)


# This implements the (single category) node selection popup when the
# tooldock is not expanded.
class CategoryPopupMenu(FramelessWindow):
    """
    A menu popup from which nodes can be dragged or clicked/activated.
    """
    triggered = Signal(QAction)
    hovered = Signal(QAction)

    def __init__(self, parent=None, **kwargs):
        # type: (Optional[QWidget], Any) -> None
        super().__init__(parent, **kwargs)
        self.setWindowFlags(self.windowFlags() | Qt.Popup)

        layout = QVBoxLayout()
        layout.setContentsMargins(6, 6, 6, 6)

        self.__menu = MenuPage()
        self.__menu.setActionRole(QtWidgetRegistry.WIDGET_ACTION_ROLE)

        if sys.platform == "darwin":
            self.__menu.view().setAttribute(Qt.WA_MacShowFocusRect, False)

        self.__menu.triggered.connect(self.__onTriggered)
        self.__menu.hovered.connect(self.hovered)

        self.__dragListener = ItemViewDragStartEventListener(self)
        self.__dragListener.dragStarted.connect(self.__onDragStarted)

        self.__menu.view().viewport().installEventFilter(self.__dragListener)
        self.__menu.view().installEventFilter(self)

        layout.addWidget(self.__menu)

        self.setLayout(layout)

        self.__action = None  # type: Optional[QAction]
        self.__loop = None    # type: Optional[QEventLoop]

    def setCategoryItem(self, item):
        """
        Set the category root item (:class:`QStandardItem`).
        """
        warnings.warn(
            "setCategoryItem is deprecated. Use the more general 'setModel'"
            "and setRootIndex", DeprecationWarning, stacklevel=2
        )
        model = item.model()
        self.__menu.setModel(model)
        self.__menu.setRootIndex(item.index())

    def setModel(self, model):
        # type: (QAbstractItemModel) -> None
        """
        Set the model.

        Parameters
        ----------
        model : QAbstractItemModel
        """
        self.__menu.setModel(model)

    def setRootIndex(self, index):
        # type: (QModelIndex) -> None
        """
        Set the root index in `model`.

        Parameters
        ----------
        index : QModelIndex
        """
        self.__menu.setRootIndex(index)

    def setActionRole(self, role):
        # type: (Qt.ItemDataRole) -> None
        """
        Set the action role in model.

        This is an item role in `model` that returns a QAction for the item.

        Parameters
        ----------
        role : Qt.ItemDataRole
        """
        self.__menu.setActionRole(role)

    def popup(self, pos=None):
        # type: (Optional[QPoint]) -> None
        """
        Show the popup at `pos`.

        Parameters
        ----------
        pos : Optional[QPoint]
            The position in global screen coordinates
        """
        if pos is None:
            pos = self.pos()
        self.adjustSize()
        geom = widget_popup_geometry(pos, self)
        self.setGeometry(geom)
        self.show()
        self.__menu.view().setFocus()

    def exec(self, pos=None):
        # type: (Optional[QPoint]) -> Optional[QAction]
        self.popup(pos)
        self.__loop = QEventLoop()

        self.__action = None
        self.__loop.exec()
        self.__loop = None

        if self.__action is not None:
            action = self.__action
        else:
            action = None
        return action

    def exec_(self, *args, **kwargs):
        warnings.warn(
            "exec_ is deprecated, use exec", DeprecationWarning, stacklevel=2
        )
        return self.exec(*args, **kwargs)

    def hideEvent(self, event):
        # type: (QHideEvent) -> None
        if self.__loop is not None:
            self.__loop.exit(0)
        super().hideEvent(event)

    def __onTriggered(self, action):
        # type: (QAction) -> None
        self.__action = action
        self.triggered.emit(action)
        self.hide()

        if self.__loop:
            self.__loop.exit(0)

    def __onDragStarted(self, index):
        # type: (QModelIndex) -> None
        desc = index.data(QtWidgetRegistry.WIDGET_DESC_ROLE)
        icon = index.data(Qt.DecorationRole)

        drag_data = QMimeData()
        drag_data.setData(
            "application/vnd.orange-canvas.registry.qualified-name",
            desc.qualified_name.encode('utf-8')
        )
        drag = QDrag(self)
        drag.setPixmap(icon.pixmap(38))
        drag.setMimeData(drag_data)

        # TODO: Should animate (accept) hide.
        self.hide()

        # When a drag is started and the menu hidden the item's tool tip
        # can still show for a short time UNDER the cursor preventing a
        # drop.
        viewport = self.__menu.view().viewport()
        filter = ToolTipEventFilter()
        viewport.installEventFilter(filter)

        drag.exec(Qt.CopyAction)

        viewport.removeEventFilter(filter)

    def eventFilter(self, obj, event):
        if isinstance(obj, QTreeView) and event.type() == QEvent.KeyPress:
            key = event.key()
            if key in [Qt.Key_Return, Qt.Key_Enter]:
                curr = obj.currentIndex()
                if curr.isValid():
                    obj.activated.emit(curr)
                    return True
        return super().eventFilter(obj, event)


class ItemViewDragStartEventListener(QObject):
    dragStarted = Signal(QModelIndex)

    def __init__(self, parent=None, **kwargs):
        # type: (Optional[QObject], Any) -> None
        super().__init__(parent, **kwargs)
        self._pos = None    # type: Optional[QPoint]
        self._index = None  # type: Optional[QPersistentModelIndex]

    def eventFilter(self, viewport, event):
        # type: (QObject, QEvent) -> bool
        view = viewport.parent()

        if event.type() == QEvent.MouseButtonPress and \
                event.button() == Qt.LeftButton:

            index = view.indexAt(event.pos())

            if index is not None:
                self._pos = event.pos()
                self._index = QPersistentModelIndex(index)

        elif event.type() == QEvent.MouseMove and self._pos is not None and \
                ((self._pos - event.pos()).manhattanLength() >=
                 QApplication.startDragDistance()):
            assert self._index is not None
            if self._index.isValid():
                # Map to a QModelIndex in the model.
                index = QModelIndex(self._index)
                self._pos = None
                self._index = None

                self.dragStarted.emit(index)

        return super().eventFilter(view, event)


class ToolTipEventFilter(QObject):
    def eventFilter(self, receiver, event):
        # type: (QObject, QEvent) -> bool
        if event.type() == QEvent.ToolTip:
            return True

        return super().eventFilter(receiver, event)


def widget_popup_geometry(pos, widget):
    # type: (QPoint, QWidget) -> QRect
    widget.ensurePolished()

    if widget.testAttribute(Qt.WA_Resized):
        size = widget.size()
    else:
        size = widget.sizeHint()

    screen = QApplication.screenAt(pos)
    if screen is None:
        screen = QApplication.primaryScreen()
    screen_geom = screen.availableGeometry()
    size = size.boundedTo(screen_geom.size())
    geom = QRect(pos, size)

    if geom.top() < screen_geom.top():
        geom.moveTop(screen_geom.top())
    if geom.left() < screen_geom.left():
        geom.moveLeft(screen_geom.left())

    bottom_margin = screen_geom.bottom() - geom.bottom()
    right_margin = screen_geom.right() - geom.right()
    if bottom_margin < 0:
        # Falls over the bottom of the screen, move it up.
        geom.translate(0, bottom_margin)

    # TODO: right to left locale
    if right_margin < 0:
        # Falls over the right screen edge, move the menu to the
        # other side of pos.
        geom.translate(-size.width(), 0)

    return geom


def popup_position_from_source(popup, source, orientation=Qt.Vertical):
    # type: (QWidget, QWidget, Qt.Orientation) -> QPoint
    popup.ensurePolished()
    source.ensurePolished()

    if popup.testAttribute(Qt.WA_Resized):
        size = popup.size()
    else:
        size = popup.sizeHint()

    screen_geom = available_screen_geometry(source)
    source_rect = QRect(source.mapToGlobal(QPoint(0, 0)), source.size())

    if orientation == Qt.Vertical:
        if source_rect.right() + size.width() < screen_geom.right():
            x = source_rect.right()
        else:
            x = source_rect.left() - size.width()

        # bottom overflow
        dy = source_rect.top() + size.height() - screen_geom.bottom()
        if dy < 0:
            y = source_rect.top()
        else:
            y = max(screen_geom.top(), source_rect.top() - dy)
    else:
        # right overflow
        dx = source_rect.left() + size.width() - screen_geom.right()
        if dx < 0:
            x = source_rect.left()
        else:
            x = max(source_rect.left() - dx, screen_geom.left())

        if source_rect.bottom() + size.height() < screen_geom.bottom():
            y = source_rect.bottom()
        else:
            y = source_rect.top() - size.height()

    return QPoint(x, y)
